from __future__ import annotations

from collections.abc import Sequence
import os
import shutil

import filelock
import pandas as pd

import audbackend
import audeer
import audformat

from audb.core import define
from audb.core import utils
from audb.core.api import cached
from audb.core.api import dependencies
from audb.core.api import latest_version
from audb.core.cache import database_cache_root
from audb.core.cache import database_tmp_root
from audb.core.cache import default_cache_root
from audb.core.dependencies import Dependencies
from audb.core.dependencies import error_message_missing_object
from audb.core.dependencies import filter_deps
from audb.core.flavor import Flavor
from audb.core.lock import FolderLock
from audb.core.utils import lookup_backend


CachedVersions = Sequence[tuple[audeer.StrictVersion, str, Dependencies]]


def _cached_versions(
    name: str,
    version: str,
    flavor: Flavor,
    cache_root: str | None,
) -> CachedVersions:
    r"""Find other cached versions of same flavor."""
    df = cached(cache_root=cache_root, name=name)
    # If no explicit cache root is given,
    # we look into the private and shared one.
    # This fixes https://github.com/audeering/audb/issues/101
    if cache_root is None and os.path.exists(default_cache_root(shared=True)):
        df = pd.concat((df, cached(name=name, shared=True)))
        # Ensure to remove duplicates,
        # which can occur if cache and shared cache
        # point to the same folder.
        # Compare https://github.com/audeering/audb/issues/314
        df = df[~df.index.duplicated(keep="first")]

    cached_versions = []
    for flavor_root, row in df.iterrows():
        if row["flavor_id"] == flavor.short_id:
            if row["version"] == version:
                continue
            deps = dependencies(
                name,
                version=row["version"],
                cache_root=cache_root,
            )
            # as it is more likely we find files
            # in newer versions, push them to front
            cached_versions.insert(
                0,
                (
                    audeer.StrictVersion(row["version"]),
                    str(flavor_root),
                    deps,
                ),
            )

    return cached_versions


def _cached_files(
    files: Sequence[str],
    deps: Dependencies,
    cached_versions: CachedVersions,
    flavor: Flavor | None,
    verbose: bool,
) -> tuple[list[str], list[str]]:
    r"""Find cached files.

    Args:
        files: media, attachment files, or table IDs
        deps: database dependencies
        cached_versions: information on cached versions of the database
        flavor: requested database flavor
        verbose: if ``True``, show progress bar

    Returns:
        ``([(<db_cache_root1>, <cached_file1>), ...)], [<missing_file1>, ...])``,
        where ``<db_cache_root1>`` is the absolute path
        to the database root,
        in which ``<cached_file1>`` is stored.
        ``<cached_file1>`` and ``<missing_file1>``
        represent the names of media, attachment files, or table IDs

    """
    cached_files = []
    missing_files = []

    for file in audeer.progress_bar(
        files,
        desc="Cached files",
        disable=not verbose,
    ):
        found = False
        if f"db.{file}.csv" in deps.tables:
            file_path = f"db.{file}.csv"
        elif f"db.{file}.parquet" in deps.tables:
            file_path = f"db.{file}.parquet"
        else:
            file_path = file
        file_version = audeer.StrictVersion(deps.version(file_path))
        for cache_version, cache_root, cache_deps in cached_versions:
            if cache_version >= file_version:
                if file_path in cache_deps:
                    if deps.checksum(file_path) == cache_deps.checksum(file_path):
                        path = os.path.join(cache_root, file_path)
                        if flavor and flavor.format is not None:
                            path = audeer.replace_file_extension(
                                path,
                                flavor.format,
                            )
                        if os.path.exists(path):
                            found = True
                            break
        if found:
            if flavor and flavor.format is not None:
                file = audeer.replace_file_extension(
                    file,
                    flavor.format,
                )
            cached_files.append((cache_root, file))
        else:
            missing_files.append(file)

    return cached_files, missing_files


def _copy_path(
    path: str,
    root_src: str,
    root_tmp: str,
    root_dst: str,
):
    r"""Copy file."""
    src_path = os.path.join(root_src, path)
    tmp_path = os.path.join(root_tmp, path)
    dst_path = os.path.join(root_dst, path)
    if os.path.isdir(src_path):
        shutil.copytree(src_path, tmp_path)
    else:
        audeer.mkdir(os.path.dirname(tmp_path))
        shutil.copy(src_path, tmp_path)
    audeer.mkdir(os.path.dirname(dst_path))
    audeer.move_file(tmp_path, dst_path)


def _database_check_complete(
    db: audformat.Database,
    db_root: str,
    flavor: Flavor,
    deps: Dependencies,
):
    def check() -> bool:
        complete = True
        for attachment in deps.attachments:
            if not os.path.exists(os.path.join(db_root, attachment)):
                return False
        for table in deps.tables:
            if not os.path.exists(os.path.join(db_root, table)):
                return False
        for media in deps.media:
            if not deps.removed(media):
                path = os.path.join(db_root, media)
                path = flavor.destination(path)
                if not os.path.exists(path):
                    return False
        return complete

    if check():
        db_root_tmp = database_tmp_root(db_root)
        db.meta["audb"]["complete"] = True
        db_original = audformat.Database.load(db_root, load_data=False)
        db_original.meta["audb"]["complete"] = True
        db_original.save(db_root_tmp, header_only=True)
        audeer.move_file(
            os.path.join(db_root_tmp, define.HEADER_FILE),
            os.path.join(db_root, define.HEADER_FILE),
        )
        audeer.rmdir(db_root_tmp)


def _database_is_complete(
    db: audformat.Database,
) -> bool:
    complete = False
    if "audb" in db.meta:
        if "complete" in db.meta["audb"]:
            complete = db.meta["audb"]["complete"]
    return complete


def _files_duration(
    db: audformat.Database,
    deps: Dependencies,
    files: Sequence[str],
    format: str | None,
):
    durs = deps().loc[files, "duration"]
    durs = durs[durs > 0]
    durs = pd.to_timedelta(durs, unit="s")
    durs.index.name = "file"
    if format is not None:
        durs.index = audformat.utils.replace_file_extension(durs.index, format)
    # Norm file path under Windows to include `\`
    if os.name == "nt":  # pragma: nocover as tested in Windows runner
        durs.index = audformat.utils.map_file_path(
            durs.index,
            os.path.normpath,
        )
    durs.index = audformat.utils.expand_file_path(durs.index, db.root)
    db._files_duration = durs.to_dict()


def _get_attachments_from_cache(
    attachments: Sequence[str],
    db_root: str,
    db: audformat.Database,
    deps: Dependencies,
    cached_versions: CachedVersions,
    flavor: Flavor,
    num_workers: int,
    verbose: bool,
) -> list[str]:
    r"""Copy files from cache.

    This function copies all files
    associated with the requested attachments
    from other cached versions
    to the new database folder.

    Args:
        attachments: sequence of attachment IDs
        db_root: database root
        db: database object
        deps: dependency object
        cached_versions: object containing information
            on existing cached versions of the database
        flavor: database flavor object
        num_workers: number of workers to use
        verbose: if ``True`` show progress bar

    Returns:
        list of attachment IDs that couldn't be found in cache

    """
    db_root_cached = [x[1] for x in cached_versions]

    paths = [db.attachments[attachment].path for attachment in attachments]

    with FolderLock(
        db_root_cached,
        timeout=define.CACHED_VERSIONS_TIMEOUT,
    ):
        cached_paths, missing_paths = _cached_files(
            paths,
            deps,
            cached_versions,
            flavor,
            verbose,
        )
        missing_attachments = [deps.archive(path) for path in missing_paths]
        db_root_tmp = database_tmp_root(db_root)

        def job(cache_root: str, file: str):
            _copy_path(file, cache_root, db_root_tmp, db_root)

        audeer.run_tasks(
            job,
            params=[([root, path], {}) for root, path in cached_paths],
            num_workers=num_workers,
            progress_bar=verbose,
            task_description="Copy attachments",
            maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
        )

        audeer.rmdir(db_root_tmp)

    return missing_attachments


def _get_files_from_cache(
    files: Sequence[str],
    files_type: str,
    db_root: str,
    deps: Dependencies,
    cached_versions: CachedVersions,
    flavor: Flavor,
    num_workers: int,
    verbose: bool,
) -> Sequence[str]:
    r"""Copy files from cache.

    This function copies requested media files
    or table files
    from other cached versions
    to the new database folder.

    Args:
        files: sequence of media files,
            attachment IDs,
            or table IDs
        files_type: ``'media'``,
            ``'table'``,
        db_root: database root
        deps: dependency object
        cached_versions: object containing information
            on existing cached versions of the database
        flavor: database flavor object
        num_workers: number of workers to use
        verbose: if ``True`` show progress bar

    Returns:
        list of files that couldn't be found in cache

    """
    db_root_cached = [x[1] for x in cached_versions]

    try:
        with FolderLock(
            db_root_cached,
            timeout=define.CACHED_VERSIONS_TIMEOUT,
        ):
            cached_files, missing_files = _cached_files(
                files,
                deps,
                cached_versions,
                flavor,
                verbose,
            )
            db_root_tmp = database_tmp_root(db_root)

            # Tables are stored as CSV or PARQUET files,
            # and are also cached as PKL files
            if files_type == "table":

                def job(cache_root: str, file: str):
                    for ext in ["csv", "parquet", "pkl"]:
                        table_file = f"db.{file}.{ext}"
                        if os.path.exists(os.path.join(cache_root, table_file)):
                            _copy_path(table_file, cache_root, db_root_tmp, db_root)

            else:

                def job(cache_root: str, file: str):
                    _copy_path(file, cache_root, db_root_tmp, db_root)

            audeer.run_tasks(
                job,
                params=[([root, file], {}) for root, file in cached_files],
                num_workers=num_workers,
                progress_bar=verbose,
                task_description=f"Copy {files_type}",
                maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
            )

            audeer.rmdir(db_root_tmp)

    except filelock.Timeout:
        missing_files = files

    return missing_files


def _get_attachments_from_backend(
    db: audformat.Database,
    attachments: Sequence[str],
    db_root: str,
    deps: Dependencies,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    r"""Load attachments from backend."""
    db_root_tmp = database_tmp_root(db_root)

    paths = [db.attachments[attachment].path for attachment in attachments]

    # create folder tree to avoid race condition
    # in os.makedirs when files are unpacked
    utils.mkdir_tree(paths, db_root_tmp)

    def job(path: str):
        archive = deps.archive(path)
        version = deps.version(path)
        archive = backend_interface.join("/", db.name, "attachment", archive + ".zip")
        backend_interface.get_archive(
            archive,
            db_root_tmp,
            version,
            tmp_root=db_root_tmp,
        )
        src_path = audeer.path(db_root_tmp, path)
        dst_path = audeer.path(db_root, path)
        audeer.mkdir(os.path.dirname(dst_path))
        audeer.move_file(
            src_path,
            dst_path,
        )

    audeer.run_tasks(
        job,
        params=[([path], {}) for path in paths],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Load attachments",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )

    audeer.rmdir(db_root_tmp)


def _get_media_from_backend(
    name: str,
    media: Sequence[str],
    db_root: str,
    flavor: Flavor | None,
    deps: Dependencies,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    r"""Load media from backend."""
    # figure out archives
    archives = set()
    archive_names = set()
    for file in media:
        archive_name = deps.archive(file)
        archive_version = deps.version(file)
        archives.add((archive_name, archive_version))
        archive_names.add(archive_name)
    # collect all files that will be extracted,
    # if we have more files than archives
    if len(deps.files) > len(deps.archives):
        files = list()
        for file in deps.media:
            archive = deps.archive(file)
            if archive in archive_names:
                files.append(file)
        media = files

    # create folder tree to avoid race condition
    # in os.makedirs when files are unpacked
    # using multi-processing
    db_root_tmp = database_tmp_root(db_root)
    utils.mkdir_tree(media, db_root)
    utils.mkdir_tree(media, db_root_tmp)

    def job(archive: str, version: str):
        archive = backend_interface.join("/", name, "media", archive + ".zip")
        # extract and move all files that are stored in the archive,
        # even if only a single file from the archive was requested
        files = backend_interface.get_archive(
            archive,
            db_root_tmp,
            version,
            tmp_root=db_root_tmp,
        )
        for file in files:
            if os.name == "nt":  # pragma: no cover
                file = file.replace(os.sep, "/")
            if flavor is not None:
                bit_depth = deps.bit_depth(file)
                channels = deps.channels(file)
                sampling_rate = deps.sampling_rate(file)
                src_path = os.path.join(db_root_tmp, file)
                file = flavor.destination(file)
                dst_path = os.path.join(db_root_tmp, file)
                try:
                    flavor(
                        src_path,
                        dst_path,
                        src_bit_depth=bit_depth,
                        src_channels=channels,
                        src_sampling_rate=sampling_rate,
                    )
                except RuntimeError:
                    raise RuntimeError(
                        f"Media file '{file}' does not support requesting a flavor."
                    )
                if src_path != dst_path:
                    os.remove(src_path)

            audeer.move_file(
                os.path.join(db_root_tmp, file),
                os.path.join(db_root, file),
            )

    audeer.run_tasks(
        job,
        params=[([archive, version], {}) for archive, version in archives],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Load media",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )

    audeer.rmdir(db_root_tmp)


def _get_tables_from_backend(
    db: audformat.Database,
    tables: Sequence[str],
    db_root: str,
    deps: Dependencies,
    backend_interface: type[audbackend.interface.Base],
    pickle_tables: bool,
    num_workers: int | None,
    verbose: bool,
):
    r"""Load tables from backend.

    Args:
        db: database
        tables: table IDs to load from backend
        db_root: database root
        deps: database dependencies
        backend_interface: backend interface
        pickle_tables: if ``True``,
            tables are cached locally
            in their original format
            and as pickle files.
            tables are stored in their original format,
            and as pickle files
            in the cache.
            This allows for faster loading,
            when loading from cache
        num_workers: number of workers
        verbose: if ``True``, show progress bar

    """
    db_root_tmp = database_tmp_root(db_root)

    def job(table: str):
        csv_file = f"db.{table}.csv"
        parquet_file = f"db.{table}.parquet"

        if csv_file in deps.tables:
            table_file = csv_file
            remote_file = backend_interface.join("/", db.name, "meta", f"{table}.zip")
            backend_interface.get_archive(
                remote_file,
                db_root_tmp,
                deps.version(table_file),
                tmp_root=db_root_tmp,
            )
        else:
            table_file = parquet_file
            remote_file = backend_interface.join(
                "/", db.name, "meta", f"{table}.parquet"
            )
            backend_interface.get_file(
                remote_file,
                os.path.join(db_root_tmp, table_file),
                deps.version(table_file),
            )

        table_files = [table_file]

        # Cache table as PKL file
        if pickle_tables:
            pickle_file = f"db.{table}.pkl"
            table_path = os.path.join(db_root_tmp, f"db.{table}")
            db[table].load(table_path)
            db[table].save(
                table_path,
                storage_format=audformat.define.TableStorageFormat.PICKLE,
            )
            table_files.append(pickle_file)

        # Move tables from tmp folder to database root
        for table_file in table_files:
            audeer.move_file(
                os.path.join(db_root_tmp, table_file),
                os.path.join(db_root, table_file),
            )

    audeer.run_tasks(
        job,
        params=[([table], {}) for table in tables],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Load tables",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )

    audeer.rmdir(db_root_tmp)


def _load_attachments(
    attachments: Sequence[str],
    backend_interface: type[audbackend.interface.Base],
    db_root: str,
    db: audformat.Database,
    version: str,
    cached_versions: CachedVersions | None,
    deps: Dependencies,
    flavor: Flavor,
    cache_root: str,
    num_workers: int,
    verbose: bool,
) -> CachedVersions | None:
    r"""Load attachments to cache.

    Args:
        attachments: list of attachment IDs
        backend_interface: backend object
        db_root: database root
        db: database object
        version: database version
        cached_versions: object representing cached versions
            of the database
        deps: database dependency object
        flavor: database flavor object
        cache_root: root path of cache
        num_workers: number of workers to use
        verbose: if ``True`` show progress bars
            for each step

    Returns:
        cached versions object
            if other versions of the database are found in cache

    """
    missing_attachments = []
    for attachment in attachments:
        path = db.attachments[attachment].path
        path = audeer.path(db_root, path)
        if not os.path.exists(path):
            missing_attachments.append(attachment)

    if missing_attachments:
        if cached_versions is None:
            cached_versions = _cached_versions(
                db.name,
                version,
                flavor,
                cache_root,
            )
        if cached_versions:
            missing_attachments = _get_attachments_from_cache(
                missing_attachments,
                db_root,
                db,
                deps,
                cached_versions,
                flavor,
                num_workers,
                verbose,
            )
        if missing_attachments:
            if backend_interface is None:
                backend_interface = lookup_backend(db.name, version)
            _get_attachments_from_backend(
                db,
                missing_attachments,
                db_root,
                deps,
                backend_interface,
                num_workers,
                verbose,
            )

    return cached_versions


def _load_files(
    files: Sequence[str],
    files_type: str,
    backend_interface: type[audbackend.interface.Base],
    db_root: str,
    db: audformat.Database,
    version: str,
    cached_versions: CachedVersions | None,
    deps: Dependencies,
    flavor: Flavor,
    cache_root: str,
    pickle_tables: bool,
    num_workers: int,
    verbose: bool,
) -> CachedVersions | None:
    r"""Load files to cache.

    Loads media files,
    attachment files,
    or table files to database root folder.

    All files not existing in cache yet
    are copied from the corresponding flavor cache
    folder of other versions of the database
    or are downloaded from the backend.

    Args:
        files: list of media files,
            attachment files,
            or table IDs
        files_type: ``'media'``,
            ``'table'``,
            or ``'attachment'``
        backend_interface: backend object
        db_root: database root
        db: database object
        version: database version
        cached_versions: object representing cached versions
            of the database
        deps: database dependency object
        flavor: database flavor object
        cache_root: root path of cache
        pickle_tables: if ``True``,
            tables are cached locally
            in their original format
            and as pickle files.
            This allows for faster loading,
            when loading from cache
        num_workers: number of workers to use
        verbose: if ``True`` show progress bars
            for each step

    Returns:
        cached versions object
            if other versions of the database are found in cache

    """
    missing_files = _missing_files(
        files,
        files_type,
        db_root,
        flavor,
        verbose,
    )
    if missing_files:
        if cached_versions is None:
            cached_versions = _cached_versions(
                db.name,
                version,
                flavor,
                cache_root,
            )
        if cached_versions:
            missing_files = _get_files_from_cache(
                missing_files,
                files_type,
                db_root,
                deps,
                cached_versions,
                flavor,
                num_workers,
                verbose,
            )
        if missing_files:
            if backend_interface is None:
                backend_interface = lookup_backend(db.name, version)
            if files_type == "media":
                _get_media_from_backend(
                    db.name,
                    missing_files,
                    db_root,
                    flavor,
                    deps,
                    backend_interface,
                    num_workers,
                    verbose,
                )
            elif files_type == "table":
                _get_tables_from_backend(
                    db,
                    missing_files,
                    db_root,
                    deps,
                    backend_interface,
                    pickle_tables,
                    num_workers,
                    verbose,
                )

    return cached_versions


def _misc_tables_used_in_scheme(
    db: audformat.Database,
) -> list[str]:
    r"""List of misc tables that are used inside a scheme.

    Args:
        db: database object

    Returns:
        unique list of misc tables used in schemes

    """
    misc_tables_used_in_scheme = []
    for scheme in db.schemes.values():
        if scheme.uses_table:
            misc_tables_used_in_scheme.append(scheme.labels)

    return audeer.unique(misc_tables_used_in_scheme)


def _misc_tables_used_in_table(
    table: audformat.Table,
) -> list[str]:
    r"""List of misc tables that are used inside schemes of a table.

    Args:
        table: table object

    Returns:
        unique list of misc tables used in schemes of the table

    """
    misc_tables_used_in_table = []
    for column_id, column in table.columns.items():
        if column.scheme_id is not None:
            scheme = table.db.schemes[column.scheme_id]
            if scheme.uses_table:
                misc_tables_used_in_table.append(scheme.labels)
    return audeer.unique(misc_tables_used_in_table)


def _missing_files(
    files: Sequence[str],
    files_type: str,
    db_root: str,
    flavor: Flavor,
    verbose: bool,
) -> list[str]:
    r"""List missing files.

    Checks for media files,
    attachment files,
    or table files
    if they exist already in database root.

    Args:
        db_root: database root
        files: list of media files,
            attachment files,
            or table IDs
        files_type: ``'media'``,
            ``'table'``,
            or ``'attachment'``
        flavor: requested database flavor
        verbose: if ``True`` show progress bar

    Returns:
        list of missing files or table IDs

    """

    def is_cached(file):
        if files_type == "table":
            path1 = os.path.join(db_root, f"db.{file}.csv")
            path2 = os.path.join(db_root, f"db.{file}.parquet")
            return os.path.exists(path1) or os.path.exists(path2)
        elif files_type == "media" and flavor.format is not None:
            # https://github.com/audeering/audb/issues/324
            cached_file = audeer.replace_file_extension(file, flavor.format)
            return os.path.exists(os.path.join(db_root, cached_file))
        else:
            return os.path.exists(os.path.join(db_root, file))

    pbar = audeer.progress_bar(files, desc=f"Missing {files_type}", disable=not verbose)
    return [file for file in pbar if not is_cached(file)]


def _remove_media(
    db: audformat.Database,
    deps: Dependencies,
    num_workers: int,
    verbose: bool,
):
    removed_files = deps.removed_media
    if removed_files:
        db.drop_files(
            removed_files,
            num_workers=num_workers,
            verbose=verbose,
        )


def _update_path(
    db: audformat.Database,
    root: str,
    full_path: bool,
    format: str | None,
    num_workers: int,
    verbose: bool,
):
    r"""Change the file path in all tables.

    Args:
        db: database object
        root: root to add to path
        full_path: if ``True`` expand file path with ``root``
        format: file extension to change to in path
        num_workers: number of workers to use
        verbose: if ``True`` show progress bar

    """
    if not full_path and format is None:
        return

    def job(table):
        if full_path:
            table._df.index = audformat.utils.expand_file_path(
                table._df.index,
                root,
            )
            # Norm file path under Windows to include `\`
            if os.name == "nt":  # pragma: nocover as tested in Windows runner
                table._df.index = audformat.utils.map_file_path(
                    table._df.index,
                    os.path.normpath,
                )
        if format is not None:
            table._df.index = audformat.utils.replace_file_extension(
                table._df.index,
                format,
            )

    tables = db.tables.values()
    audeer.run_tasks(
        job,
        params=[([table], {}) for table in tables],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Update file path",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )


def filtered_dependencies(
    name: str,
    version: str,
    media: str | Sequence[str],
    tables: str | Sequence[str],
    cache_root: str = None,
) -> pd.DataFrame:
    r"""Filter media by tables.

    Return all media files from ``media``
    that are referenced in at least one table
    from ``tables``.
    This will download all tables.

    Args:
        name: name of database
        version: version of database
        media: media files
        tables: table IDs
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        filtered dependencies

    """
    deps = dependencies(name, version=version, cache_root=cache_root)
    if tables is None and media is None:
        df = deps()
    else:
        # Load header to get list of tables
        db = load_header(name, version=version, cache_root=cache_root)
        tables = filter_deps(tables, list(db), "table")
        tables = [t for t in tables if t not in list(db.misc_tables)]
        # Gather media files from tables
        available_media = []
        for table in tables:
            df = load_table(
                name,
                table,
                version=version,
                cache_root=cache_root,
                verbose=False,
            )
            available_media += list(df.index.get_level_values("file").unique())

        if len(available_media) > 0:
            media = filter_deps(media, deps.media, "media", name, version)
            available_media = [m for m in media if m in list(set(available_media))]
        df = deps().loc[available_media]

    return df


def load(
    name: str,
    *,
    version: str = None,
    only_metadata: bool = False,
    bit_depth: int = None,
    channels: int | Sequence[int] = None,
    format: str = None,
    mixdown: bool = False,
    sampling_rate: int = None,
    attachments: str | Sequence[str] = None,
    tables: str | Sequence[str] = None,
    media: str | Sequence[str] = None,
    removed_media: bool = False,
    full_path: bool = True,
    pickle_tables: bool = True,
    cache_root: str = None,
    num_workers: int | None = 1,
    timeout: float = -1,
    verbose: bool = True,
) -> audformat.Database | None:
    r"""Load database.

    Loads meta and media files of a database to the local cache and returns
    a :class:`audformat.Database` object.

    By setting
    ``bit_depth``,
    ``channels``,
    ``format``,
    ``mixdown``,
    and ``sampling_rate``
    we can request a specific flavor of the database.
    In that case media files are automatically converted to the desired
    properties (see also :class:`audb.Flavor`).

    It is possible to filter meta and media files with the arguments
    ``tables`` and ``media``.
    Only media files with at least one reference are loaded.
    I.e. filtering meta files, may also remove media files.
    Likewise, references to missing media files will be removed, too.
    I.e. filtering media files, may also remove entries from the meta files.

    Args:
        name: name of database
        version: version string, latest if ``None``
        only_metadata: load only header and tables of database
        bit_depth: bit depth, one of ``16``, ``24``, ``32``
        channels: channel selection, see :func:`audresample.remix`.
            Note that media files with too few channels
            will be first upsampled by repeating the existing channels.
            E.g. ``channels=[0, 1]`` upsamples all mono files to stereo,
            and ``channels=[1]`` returns the second channel
            of all multi-channel files
            and all mono files
        format: file format, one of ``'flac'``, ``'wav'``
        mixdown: apply mono mix-down
        sampling_rate: sampling rate in Hz, one of
            ``8000``, ``16000``, ``22050``, ``24000``, ``44100``, ``48000``
        attachments: load only attachment files
            for the attachments
            matching the regular expression
            or provided in the list.
            If set to ``[]`` no attachments are loaded
        tables: load only tables and misc tables
            matching the regular expression
            or provided in the list.
            Media files not referenced
            in the selected tables
            are automatically excluded, too.
            If set to ``[]``
            no tables and media files are loaded.
            Misc tables used in schemes are always loaded
        media: load only media files
            matching the regular expression
            or provided in the list.
            Excluded media files are
            automatically removed from the tables, too.
            This may result in empty tables.
            If set to ``[]``
            no media files are loaded
            and all tables except
            misc tables will be empty
        removed_media: keep rows that reference removed media
        full_path: replace relative with absolute file paths
        pickle_tables: if ``True``,
            tables are cached locally
            in their original format
            and as pickle files.
            This allows for faster loading,
            when loading from cache
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5
        timeout: maximum wait time if another thread or process is already
            accessing the database. If timeout is reached, ``None`` is
            returned. If timeout < 0 the method will block until the
            database can be accessed
        verbose: show debug messages

    Returns:
        database object

    Raises:
        ValueError: if attachment, table or media is requested
            that is not part of the database
        ValueError: if a non-supported ``bit_depth``,
            ``format``,
            or ``sampling_rate``
            is requested
        RuntimeError: if a flavor is requested,
            but the database contains media files,
            that don't contain audio,
            e.g. text files

    Examples:
        >>> db = audb.load(
        ...     "emodb",
        ...     version="1.4.1",
        ...     tables=["emotion", "files"],
        ...     only_metadata=True,
        ...     full_path=False,
        ...     verbose=False,
        ... )
        >>> list(db.tables)
        ['emotion', 'files']

    """
    if version is None:
        version = latest_version(name)

    db = None
    cached_versions = None
    flavor = Flavor(
        channels=channels,
        format=format,
        mixdown=mixdown,
        bit_depth=bit_depth,
        sampling_rate=sampling_rate,
    )
    db_root = database_cache_root(name, version, cache_root, flavor)

    if verbose:  # pragma: no cover
        print(f"Get:   {name} v{version}")
        print(f"Cache: {db_root}")

    deps = dependencies(
        name,
        version=version,
        cache_root=cache_root,
        verbose=verbose,
    )

    try:
        with FolderLock(db_root, timeout=timeout):
            # Start with database header without tables
            db, backend_interface = load_header_to(
                db_root,
                name,
                version,
                flavor=flavor,
                add_audb_meta=True,
            )

            db_is_complete = _database_is_complete(db)

            # load attachments
            if not db_is_complete and not only_metadata:
                # filter attachments
                requested_attachments = filter_deps(
                    attachments,
                    db.attachments,
                    "attachment",
                )

                cached_versions = _load_attachments(
                    requested_attachments,
                    backend_interface,
                    db_root,
                    db,
                    version,
                    cached_versions,
                    deps,
                    flavor,
                    cache_root,
                    num_workers,
                    verbose,
                )

            # filter tables (convert regexp pattern to list of tables)
            requested_tables = filter_deps(tables, list(db), "table")

            # add/split into misc tables used in a scheme
            # and all other (misc) tables
            requested_misc_tables = _misc_tables_used_in_scheme(db)
            requested_tables = [
                table
                for table in requested_tables
                if table not in requested_misc_tables
            ]

            # load missing tables
            if not db_is_complete:
                for _tables in [
                    requested_misc_tables,
                    requested_tables,
                ]:
                    # need to load misc tables used in a scheme first
                    # as loading is done in parallel
                    cached_versions = _load_files(
                        _tables,
                        "table",
                        backend_interface,
                        db_root,
                        db,
                        version,
                        cached_versions,
                        deps,
                        flavor,
                        cache_root,
                        pickle_tables,
                        num_workers,
                        verbose,
                    )
            requested_tables = requested_misc_tables + requested_tables

            # filter tables
            if tables is not None:
                db.pick_tables(requested_tables)

            # load tables
            for table in requested_tables:
                db[table].load(os.path.join(db_root, f"db.{table}"))

            # filter media
            requested_media = filter_deps(
                media,
                db.files,
                "media",
                name,
                version,
            )

            # load missing media
            if not db_is_complete and not only_metadata:
                cached_versions = _load_files(
                    requested_media,
                    "media",
                    backend_interface,
                    db_root,
                    db,
                    version,
                    cached_versions,
                    deps,
                    flavor,
                    cache_root,
                    False,
                    num_workers,
                    verbose,
                )

            # filter media
            if media is not None or tables is not None:
                db.pick_files(requested_media)

            if not removed_media:
                _remove_media(db, deps, num_workers, verbose)

            # Adjust full paths and file extensions in tables
            _update_path(
                db,
                db_root,
                full_path,
                flavor.format,
                num_workers,
                verbose,
            )

            # set file durations
            _files_duration(
                db,
                deps,
                requested_media,
                flavor.format,
            )

            # check if database is now complete
            if not db_is_complete:
                _database_check_complete(
                    db,
                    db_root,
                    flavor,
                    deps,
                )

    except filelock.Timeout:
        utils.timeout_warning()

    return db


def load_attachment(
    name: str,
    attachment: str,
    *,
    version: str = None,
    cache_root: str = None,
    verbose: bool = True,
) -> list[str]:
    r"""Load attachment(s) of database.

    Args:
        name: name of database
        attachment: attachment ID to load
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used
        verbose: show debug messages

    Returns:
        list of file paths belonging to attachment

    Raises:
        ValueError: if an attachment ID is requested
            that is not part of the database

    Examples:
        >>> paths = audb.load_attachment(
        ...     "emodb",
        ...     "bibtex",
        ...     version="1.4.1",
        ...     verbose=False,
        ... )
        >>> os.path.basename(paths[0])
        'burkhardt2005emodb.bib'

    """
    if version is None:
        version = latest_version(name)

    db_root = database_cache_root(name, version, cache_root)

    if verbose:  # pragma: no cover
        print(f"Get:   {name} v{version}")
        print(f"Cache: {db_root}")

    deps = dependencies(
        name,
        version=version,
        cache_root=cache_root,
        verbose=verbose,
    )

    if attachment not in deps.archives:
        msg = error_message_missing_object(
            "attachment",
            [attachment],
            name,
            version,
        )
        raise ValueError(msg)

    with FolderLock(db_root):
        # Start with database header
        db, backend_interface = load_header_to(
            db_root,
            name,
            version,
        )

        # Load attachment
        _load_attachments(
            [attachment],
            backend_interface,
            db_root,
            db,
            version,
            None,
            deps,
            Flavor(),
            cache_root,
            1,
            verbose,
        )

    attachment_files = db.attachments[attachment].files
    attachment_files = [
        os.path.join(db_root, os.path.normpath(file))  # convert "/" to os.sep
        for file in attachment_files
    ]
    return attachment_files


def load_header(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> audformat.Database:
    r"""Load header of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        database object without table data

    """
    if version is None:
        version = latest_version(name)

    db_root = database_cache_root(name, version, cache_root)

    with FolderLock(db_root):
        db, _ = load_header_to(db_root, name, version)

    return db


def load_header_to(
    db_root: str,
    name: str,
    version: str,
    *,
    flavor: Flavor = None,
    add_audb_meta: bool = False,
    overwrite: bool = False,
) -> tuple[audformat.Database, type[audbackend.interface.Base] | None]:
    r"""Load database header from folder or backend.

    If the database header cannot be found in ``db_root``
    it will search for the backend that contains the database,
    load it from there,
    and store it in ``db_root``.

    Args:
        db_root: folder of database
        name: name of database
        version: version of database
        flavor: flavor of database,
            needed if ``add_audb_meta`` is True
        add_audb_meta: if ``True`` it adds an ``audb`` meta entry
            to the database header before storing it in cache
        overwrite: always load header from backend
            and overwrite the one found in ``db_root``

    Returns:
        database header and backend

    """
    backend_interface = None
    local_header = os.path.join(db_root, define.HEADER_FILE)
    if overwrite or not os.path.exists(local_header):
        backend_interface = lookup_backend(name, version)
        remote_header = backend_interface.join("/", name, define.HEADER_FILE)
        if add_audb_meta:
            db_root_tmp = database_tmp_root(db_root)
            local_header = os.path.join(db_root_tmp, define.HEADER_FILE)
        backend_interface.get_file(remote_header, local_header, version)
        if add_audb_meta:
            db = audformat.Database.load(db_root_tmp, load_data=False)
            db.meta["audb"] = {
                "root": db_root,
                "version": version,
                "flavor": flavor.arguments,
                "complete": False,
            }
            db.save(db_root_tmp, header_only=True)
            audeer.move_file(
                os.path.join(db_root_tmp, define.HEADER_FILE),
                os.path.join(db_root, define.HEADER_FILE),
            )
            audeer.rmdir(db_root_tmp)

    return audformat.Database.load(db_root, load_data=False), backend_interface


def load_media(
    name: str,
    media: str | Sequence[str],
    *,
    version: str = None,
    bit_depth: int = None,
    channels: int | Sequence[int] = None,
    format: str = None,
    mixdown: bool = False,
    sampling_rate: int = None,
    cache_root: str = None,
    num_workers: int | None = 1,
    timeout: float = -1,
    verbose: bool = True,
) -> list | None:
    r"""Load media file(s).

    If you are interested in media files
    and not the corresponding tables,
    you can use :func:`audb.load_media`
    to load them.
    This will not download any table files
    to your disk,
    but share the cache with :func:`audb.load`.

    Args:
        name: name of database
        media: load media files provided in the list
        version: version of database
        bit_depth: bit depth, one of ``16``, ``24``, ``32``
        channels: channel selection, see :func:`audresample.remix`.
            Note that media files with too few channels
            will be first upsampled by repeating the existing channels.
            E.g. ``channels=[0, 1]`` upsamples all mono files to stereo,
            and ``channels=[1]`` returns the second channel
            of all multi-channel files
            and all mono files
        format: file format, one of ``'flac'``, ``'wav'``
        mixdown: apply mono mix-down
        sampling_rate: sampling rate in Hz, one of
            ``8000``, ``16000``, ``22050``, ``24000``, ``44100``, ``48000``
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5
        timeout: maximum wait time if another thread or process is already
            accessing the database. If timeout is reached, ``None`` is
            returned. If timeout < 0 the method will block until the
            database can be accessed
        verbose: show debug messages

    Returns:
        paths to media files

    Raises:
        ValueError: if a media file is requested
            that is not part of the database
        ValueError: if a non-supported ``bit_depth``,
            ``format``,
            or ``sampling_rate``
            is requested

    Examples:
        >>> paths = audb.load_media(
        ...     "emodb",
        ...     ["wav/03a01Fa.wav"],
        ...     version="1.4.1",
        ...     format="flac",
        ...     verbose=False,
        ... )
        >>> paths[0].split(os.path.sep)[-5:]
        ['emodb', '1.4.1', '40bb2241', 'wav', '03a01Fa.flac']

    """
    media = audeer.to_list(media)
    if len(media) == 0:
        return []

    if version is None:
        version = latest_version(name)

    files = None
    flavor = Flavor(
        channels=channels,
        format=format,
        mixdown=mixdown,
        bit_depth=bit_depth,
        sampling_rate=sampling_rate,
    )
    db_root = database_cache_root(name, version, cache_root, flavor)

    if verbose:  # pragma: no cover
        print(f"Get:   {name} v{version}")
        print(f"Cache: {db_root}")

    deps = dependencies(
        name,
        version=version,
        cache_root=cache_root,
        verbose=verbose,
    )

    available_files = deps.media
    for media_file in media:
        if media_file not in available_files:
            msg = error_message_missing_object(
                "media",
                [media_file],
                name,
                version,
            )
            raise ValueError(msg)

    try:
        with FolderLock(db_root, timeout=timeout):
            # Start with database header without tables
            db, backend_interface = load_header_to(
                db_root,
                name,
                version,
                flavor=flavor,
                add_audb_meta=True,
            )

            db_is_complete = _database_is_complete(db)

            # load missing media
            if not db_is_complete:
                _load_files(
                    media,
                    "media",
                    backend_interface,
                    db_root,
                    db,
                    version,
                    None,
                    deps,
                    flavor,
                    cache_root,
                    False,
                    num_workers,
                    verbose,
                )

            if format is not None:
                media = [audeer.replace_file_extension(m, format) for m in media]
            files = [
                os.path.join(db_root, os.path.normpath(file))  # convert "/" to os.sep
                for file in media
            ]

    except filelock.Timeout:
        utils.timeout_warning()

    return files


def load_table(
    name: str,
    table: str,
    *,
    version: str = None,
    map: dict[str, str | Sequence[str]] = None,
    pickle_tables: bool = True,
    cache_root: str = None,
    num_workers: int | None = 1,
    verbose: bool = True,
) -> pd.DataFrame:
    r"""Load a database table.

    If you are interested in a single table
    from a database
    you can use :func:`audb.load_table`
    to directly load it.
    This will not download any media files
    to your disk,
    but share the cache with :func:`audb.load`.

    Args:
        name: name of database
        table: load table from database
        version: version of database
        map: map scheme or scheme fields to column values.
            For example if your table holds a column ``speaker`` with
            speaker IDs, which is assigned to a scheme that contains a
            dict mapping speaker IDs to age and gender entries,
            ``map={'speaker': ['age', 'gender']}``
            will replace the column with two new columns that map ID
            values to age and gender, respectively.
            To also keep the original column with speaker IDS, you can do
            ``map={'speaker': ['speaker', 'age', 'gender']}``
        pickle_tables: if ``True``,
            tables are cached locally
            in their original format
            and as pickle files.
            This allows for faster loading,
            when loading from cache
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5
        verbose: show debug messages

    Returns:
        database table

    Raises:
        ValueError: if a table is requested
            that is not part of the database

    Examples:
        >>> df = audb.load_table("emodb", "emotion", version="1.4.1", verbose=False)
        >>> df[:3]
                           emotion  emotion.confidence
        file
        wav/03a01Fa.wav  happiness                0.90
        wav/03a01Nc.wav    neutral                1.00
        wav/03a01Wa.wav      anger                0.95
        >>> df = audb.load_table("emodb", "files", version="1.4.1", verbose=False)
        >>> df[:3]
                                         duration speaker transcription
        file
        wav/03a01Fa.wav    0 days 00:00:01.898250       3           a01
        wav/03a01Nc.wav    0 days 00:00:01.611250       3           a01
        wav/03a01Wa.wav 0 days 00:00:01.877812500       3           a01
        >>> df = audb.load_table(
        ...     "emodb",
        ...     "files",
        ...     version="1.4.1",
        ...     map={"speaker": "age"},
        ...     verbose=False,
        ... )
        >>> df[:3]
                                         duration transcription  age
        file
        wav/03a01Fa.wav    0 days 00:00:01.898250           a01   31
        wav/03a01Nc.wav    0 days 00:00:01.611250           a01   31
        wav/03a01Wa.wav 0 days 00:00:01.877812500           a01   31

    """
    if version is None:
        version = latest_version(name)

    db_root = database_cache_root(name, version, cache_root)

    if verbose:  # pragma: no cover
        print(f"Get:   {name} v{version}")
        print(f"Cache: {db_root}")

    deps = dependencies(
        name,
        version=version,
        cache_root=cache_root,
        verbose=verbose,
    )

    if table not in deps.table_ids:
        msg = error_message_missing_object(
            "table",
            [table],
            name,
            version,
        )
        raise ValueError(msg)

    with FolderLock(db_root):
        # Start with database header without tables
        db, backend_interface = load_header_to(
            db_root,
            name,
            version,
        )

        # Find only those misc tables used in schemes of the requested table
        scheme_misc_tables = []
        for column_id, column in db[table].columns.items():
            if column.scheme_id is not None:
                scheme = db.schemes[column.scheme_id]
                if scheme.uses_table:
                    scheme_misc_tables.append(scheme.labels)
        scheme_misc_tables = audeer.unique(scheme_misc_tables)

        # Load table
        tables = scheme_misc_tables + [table]
        for _table in tables:
            table_file = os.path.join(db_root, f"db.{_table}")
            # `_load_files()` downloads a table
            # from the backend,
            # if it cannot find its corresponding csv or parquet file
            if not os.path.exists(f"{table_file}.pkl"):
                _load_files(
                    [_table],
                    "table",
                    backend_interface,
                    db_root,
                    db,
                    version,
                    None,
                    deps,
                    Flavor(),
                    cache_root,
                    pickle_tables,
                    num_workers,
                    verbose,
                )
            db[_table].load(table_file)

    if map is None:
        df = db[table]._df
    else:
        df = db[table].get(map=map)

    return df



import pandas as pd
import tabulate

import audeer


cache = audeer.mkdir("./cache")
engines = ["", "_polars"]


def parse_df(fp, engine):
    df = pd.read_csv(fp, index_col="method")
    df["engine"] = engine
    if engine == "pandas":
        df.drop(columns=["string", "object"], inplace=True)

    df.rename(columns={"polars": "t", "pyarrow": "t"}, inplace=True)
    return df


def get_col(sr):
    """Return columns to annotate df."""
    name = sr.idxmin()
    return [name, sr.max() / sr.min()]


df_polars = parse_df(audeer.path(cache, "results_polars.csv"), "polars")
df_pandas = parse_df(audeer.path(cache, "results.csv"), "pandas")
df = pd.concat([df_pandas, df_polars], axis=0)

df = df.reset_index().pivot(
    index="method",
    columns="engine",
    values="t",
)


df[["winner", "factor"]] = df[["pandas", "polars"]].apply(
    lambda x: get_col(x), axis=1, result_type="expand"
)


table = tabulate.tabulate(df, headers="keys", tablefmt="github", floatfmt=".3f")
print(table)



import pytest
import yaml

import audeer

import audb


@pytest.fixture()
def config_files(tmpdir, request):
    """Provide user config files.

    The config file at ``.config/audb.yaml``
    sets ``cache_root`` to ``~/user1``.
    The config file at ``.audb.yam``
    sets ``cache_root`` to ``~/user2``.

    Args:
        tmpdir: tmpdir fixture.
            The tmpdir is used as the home folder
            for storing user config files
        request: request fixture
            for selecting which user config file to create.
            ``"default"`` will create a file at ``tmpdir/.config/audb.yaml``,
            ``"deprecated"`` will create a file at ``tmpdir/.audb.yaml``,
            ``"both"`` will create both files

    """
    home = audeer.mkdir(tmpdir)
    current_user_config_file = audb.core.define.USER_CONFIG_FILE
    current_deprecated_user_config_file = audb.core.define.DEPRECATED_USER_CONFIG_FILE
    audb.core.config.USER_CONFIG_FILE = audeer.path(home, ".config", "audb.yaml")
    audb.core.config.DEPRECATED_USER_CONFIG_FILE = audeer.path(home, ".audb.yaml")
    if request.param in ["both", "default"]:
        audeer.mkdir(home, ".config")
        with open(audeer.path(home, ".config", "audb.yaml"), "w") as fp:
            fp.write("cache_root: ~/user1\n")
    if request.param in ["both", "deprecated"]:
        with open(audeer.path(home, ".audb.yaml"), "w") as fp:
            fp.write("cache_root: ~/user2\n")

    yield

    audb.core.config.USER_CONFIG_FILE = current_user_config_file
    audb.core.config.DEPRECATED_USER_CONFIG_FILE = current_deprecated_user_config_file


def test_config_file(tmpdir):
    root = audeer.mkdir(tmpdir)

    config_file = audeer.path(root, ".audb.yaml")

    # Try loading non-existing file
    config = audb.core.config.load_configuration_file(config_file)
    assert config == {}

    # Add a custom cache entry
    # and check if combining with global config works
    with open(config_file, "w") as cf:
        cf.write("cache_root: ~/user\n")

    config = audb.core.config.load_configuration_file(config_file)
    assert config == {"cache_root": "~/user"}

    global_config = audb.core.config.load_configuration_file(
        audb.core.config.global_config_file
    )
    global_config.update(config)
    assert global_config["cache_root"] == "~/user"
    assert global_config["shared_cache_root"] == "/data/audb"

    # Fail for wrong repositories entries
    with open(config_file, "w") as cf:
        cf.write("repositories:\n")
    error_msg = (
        "You cannot specify an empty 'repositories:' section "
        f"in the configuration file '{audb.core.define.USER_CONFIG_FILE}'."
    )
    with pytest.raises(ValueError, match=error_msg):
        audb.core.config.load_configuration_file(config_file)

    with open(config_file, "w") as cf:
        cf.write("repositories:\n")
    error_msg = "You cannot specify an empty 'repositories:' section"
    with pytest.raises(ValueError, match=error_msg):
        audb.core.config.load_configuration_file(config_file)

    with open(config_file, "w") as cf:
        cf.write("repositories:\n")
        cf.write("  - host: some-host\n")
        cf.write("    backend: some-backend\n")
    error_msg = "Your repository is missing a 'name' entry"
    with pytest.raises(ValueError, match=error_msg):
        audb.core.config.load_configuration_file(config_file)

    with open(config_file, "w") as cf:
        cf.write("repositories:\n")
        cf.write("  - name: my-repo\n")
    error_msg = "Your repository is missing a 'host' entry."
    with pytest.raises(ValueError, match=error_msg):
        audb.core.config.load_configuration_file(config_file)

    with open(config_file, "w") as cf:
        cf.write("repositories:\n")
        cf.write("  - name: my-repo\n")
        cf.write("    host: some-host\n")
    error_msg = "Your repository is missing a 'backend' entry"
    with pytest.raises(ValueError, match=error_msg):
        audb.core.config.load_configuration_file(config_file)

    # Load custom repository
    with open(config_file, "w") as cf:
        cf.write("repositories:\n")
        cf.write("  - name: my-repo\n")
        cf.write("    host: some-host\n")
        cf.write("    backend: some-backend\n")
    config = audb.core.config.load_configuration_file(config_file)
    assert config == {
        "repositories": [
            {
                "name": "my-repo",
                "host": "some-host",
                "backend": "some-backend",
            },
        ]
    }


@pytest.mark.parametrize(
    "config_files, expected",
    [
        ("both", "~/user1"),
        ("default", "~/user1"),
        ("deprecated", "~/user2"),
    ],
    indirect=["config_files"],
)
def test_deprecated_config_file(config_files, expected):
    """Test loading of user configuration files.

    It especially checks,
    if the default configuration file
    overwrites a deprecated user config file.

    Args:
        config_files: config_files fixture
        expected: expected value for ``cache_root``
            config entry

    """
    config = audb.core.config.load_config()
    assert config["cache_root"] == expected


def test_empty_config_file(tmp_path):
    """Test loading an empty config file."""
    empty_config = tmp_path / "empty.yaml"
    empty_config.write_text("")
    config = audb.core.config.load_configuration_file(empty_config)
    assert config == {}


def test_invalid_config_file(tmp_path):
    """Test loading a broken config file."""
    invalid_config = tmp_path / "invalid.yaml"
    invalid_config.write_text("{invalid: yaml: content}")
    with pytest.raises(yaml.YAMLError):
        audb.core.config.load_configuration_file(invalid_config)



from doctest import ELLIPSIS
from doctest import NORMALIZE_WHITESPACE
import os

import pytest
import sybil
from sybil.parsers.rest import DocTestParser

import audb


def imports(namespace):
    """Provide Python modules to namespace."""
    namespace["audb"] = audb


@pytest.fixture(scope="module", autouse=True)
def cache(tmpdir_factory):
    r"""Provide a reusable cache for docstring tests.

    As we rely on emodb from the public repo,
    it makes sense to cache it
    across all docstring tests.

    """
    cache = tmpdir_factory.mktemp("cache")
    # We use the environment variable here
    # to ensure audb.config.CACHE_ROOT
    # does still return the default config value
    # in the doctest
    env_cache = os.environ.get("AUDB_CACHE_ROOT", None)
    env_shared_cache = os.environ.get("AUDB_SHARED_CACHE_ROOT", None)
    os.environ["AUDB_CACHE_ROOT"] = str(cache)
    os.environ["AUDB_SHARED_CACHE_ROOT"] = str(cache)

    yield

    if env_cache is None:
        del os.environ["AUDB_CACHE_ROOT"]
    else:  # pragma: nocover
        os.environ["AUDB_CACHE_ROOT"] = env_cache

    if env_shared_cache is None:
        del os.environ["AUDB_SHARED_CACHE_ROOT"]
    else:  # pragma: nocover
        os.environ["AUDB_SHARED_CACHE_ROOT"] = env_shared_cache


@pytest.fixture(autouse=True)
def public_repository():
    r"""Provide access to the public Artifactory repository."""
    audb.config.REPOSITORIES = [
        audb.Repository(
            name="audb-public",
            host="s3.dualstack.eu-north-1.amazonaws.com",
            backend="s3",
        ),
    ]

    yield

    # Remove public repo
    audb.config.REPOSITORIES.pop()


# Collect doctests
pytest_collect_file = sybil.Sybil(
    parsers=[DocTestParser(optionflags=ELLIPSIS + NORMALIZE_WHITESPACE)],
    patterns=["*.py"],
    fixtures=[
        "cache",
        "public_repository",
    ],
    setup=imports,
).pytest()



from collections.abc import Sequence
import os

import numpy as np
import pyarrow.parquet as parquet
import pytest

import audeer
import audformat
import audiofile

import audb


@pytest.fixture
def build_dir(tmpdir) -> str:
    r"""Build dir fixture.

    Args:
        tmpdir: tmpdir fixture

    Returns:
        path to build dir

    """
    yield audeer.mkdir(tmpdir, "build")


@pytest.fixture
def db(build_dir) -> str:
    r"""Database fixture.

    Creates a minimal database,
    containing a media file
    and a table.
    The media file is stored in the provided ``build_dir``,
    but the database is not yet stored to disk,
    enabling to select the storage format
    of the tables inside a test function.

    Args:
        build_dir: build_dir fixture

    Returns:
        database object

    """
    name = "test-db"
    file = "data/file1.wav"
    table = "files"
    data_dir = audeer.mkdir(build_dir, os.path.dirname(os.path.normpath(file)))
    audio_file = audeer.path(data_dir, os.path.basename(os.path.normpath(file)))
    signal = np.zeros((2, 1000))
    sampling_rate = 8000
    audiofile.write(audio_file, signal, sampling_rate)
    db = audformat.Database(name)
    db.schemes["speaker"] = audformat.Scheme("str")
    index = audformat.filewise_index([file])
    db[table] = audformat.Table(index)
    db[table]["speaker"] = audformat.Column(scheme_id="speaker")
    db[table]["speaker"].set(["adam"])

    yield db


def assert_db_published_to_repo(
    db: audformat.Database,
    deps: audb.Dependencies,
    repository: audb.Repository,
    version: str,
    storage_format: str,
    *,
    table_changed: bool = True,
):
    r"""Assert files are published to repository.

    Args:
        db: database object,
            used to collect all media files and tables
        deps: dependency table,
            used to get name of media file archives
        repository: repository the database was published to
        version: version of database
        storage_format: table storage format on repository,
            ``"csv"`` or ``"parquet"``
        table_changed: if ``True``,
            it assumes the tables have changed
            between database versions

    """
    repo = audeer.path(repository.host, repository.name)

    dependency_file = "db.parquet"
    header_file = "db.yaml"
    files = list(db.files)
    tables = list(db)
    archives = [f"{deps.archive(file)}.zip" for file in files]
    if storage_format == "csv":
        ext = "zip"
    elif storage_format == "parquet":
        ext = "parquet"
    meta_files = [f"{table}.{ext}" for table in tables]

    def repo_path(*args):
        return audeer.path(repo, db.name, *args)

    expected_paths = [
        repo_path("1.0.0", dependency_file),
        repo_path("1.0.0", header_file),
    ]
    if version == "1.1.0":
        expected_paths.append(repo_path("1.1.0", dependency_file))
        expected_paths.append(repo_path("1.1.0", header_file))
    for archive in archives:
        expected_paths.append(repo_path("media", "1.0.0", archive))
    for meta_file in meta_files:
        expected_paths.append(repo_path("meta", "1.0.0", meta_file))
    if version == "1.1.0" and table_changed:
        for meta_file in meta_files:
            expected_paths.append(repo_path("meta", "1.1.0", meta_file))

    assert audeer.list_file_names(repo, recursive=True) == expected_paths


def assert_db_saved_to_dir(
    db: audformat.Database,
    root: str,
    storage_formats: Sequence[str],
):
    r"""Assert all database files are stored to the build dir.

    Args:
        db: database
        root: path to folder the database saved to
        storage_formats: storage formats,
            the tables have been stored to the build folder

    """
    other_storage_formats = [
        storage_format
        for storage_format in ["csv", "parquet", "pkl"]
        if storage_format not in storage_formats
    ]
    tables = list(db)
    for storage_format in storage_formats:
        for table in tables:
            table_file = f"db.{table}.{storage_format}"
            assert os.path.exists(os.path.join(root, table_file))
    for storage_format in other_storage_formats:
        for table in tables:
            table_file = f"db.{table}.{storage_format}"
            assert not os.path.exists(os.path.join(root, table_file))
    for media_file in list(db.files):
        assert os.path.exists(os.path.join(root, media_file))


def assert_dependency_table(
    db: audformat.Database,
    build_dir: str,
    deps: audb.Dependencies,
    storage_format: str,
):
    r"""Assert dependency table entries.

    Args:
        db: database
        build_dir: build dir the database was published from
        deps: dependency table
        storage_format: storage format of database tables

    """
    tables = list(db)
    assert deps.tables == [f"db.{table}.{storage_format}" for table in tables]
    assert deps.media == list(db.files)
    for table in list(db):
        table_file = f"db.{table}.{storage_format}"
        assert deps.checksum(table_file) == expected_table_checksum(
            audeer.path(build_dir, table_file)
        )
        if storage_format == "csv":
            assert deps.archive(table_file) == table
        elif storage_format == "parquet":
            assert deps.archive(table_file) == ""


def expected_table_checksum(path: str) -> str:
    r"""Expected checksum of table file.

    MD5 sum for CSV table,
    ``"hash"`` metadata entry for PARQUET files.

    Args:
        path: path to table file

    """
    if audeer.file_extension(path) == "parquet":
        # See https://github.com/audeering/audformat/pull/419
        return parquet.read_schema(path).metadata[b"hash"].decode()
    else:
        return audeer.md5(path)


def publish_db(build_dir, db, repository, version, storage_format):
    r"""Publish database.

    Args:
        build_dir: build dir
        db: database
        repository: repository to publish database to
        version: version of database
        storage_format: storage format of database tables

    Returns:
        dependency table of published database

    """
    db.save(build_dir, storage_format=storage_format)
    return audb.publish(build_dir, version, repository)


def update_db(db: audformat.Database) -> audformat.Database:
    r"""Update database by adding a column to all tables.

    Args:
        db: database

    Returns:
        updated database

    """
    for table in list(db):
        db[table]["object"] = audformat.Column()
        db[table]["object"].set(["!!!"])
    return db


@pytest.mark.parametrize("storage_format", ["csv", "parquet"])
class TestPublishTableStorageFormat:
    r"""Test publishing and of tables for different storage formats.

    Tables are stored as CSV or PARQUET files.
    ``audb`` is handling them slightly different:

    * CSV tables are published as ``<table-id>.zip``,
      and the table checksum is calculated
      on the original CSV file
    * PARQUET tables are published as ``<table-id>.parquet``
      and the table checksum is read
      from the header of the PARQUET file

    Args:
        storage_format: storage format of tables

    """

    def test_database_save(self, build_dir, db, storage_format):
        r"""Test correct files are stored to build dir.

        Args:
            build_dir: build dir fixture
            db: database fixture
            storage_format: storage format of database tables

        """
        db.save(build_dir, storage_format=storage_format)
        assert_db_saved_to_dir(db, build_dir, [storage_format])

    def test_database_publish(
        self,
        build_dir: str,
        db: audformat.Database,
        repository: audb.Repository,
        storage_format: str,
    ):
        r"""Tests files stored in repository and dependency table entries.

        Tables are stored differently in the repository
        for different storage formats,
        and their dependency table entry looks different as well.

        Args:
            build_dir: build dir fixture
            db: database fixture
            repository: repository fixture
            storage_format: database table storage format

        """
        version = "1.0.0"
        deps = publish_db(build_dir, db, repository, version, storage_format)

        assert_db_published_to_repo(db, deps, repository, version, storage_format)
        assert_dependency_table(db, build_dir, deps, storage_format)

    def test_database_load(
        self,
        build_dir: str,
        db: audformat.Database,
        repository: audb.Repository,
        storage_format: str,
    ):
        r"""Test correct files in cache after loading database.

        Args:
            build_dir: build dir fixture
            db: database fixture
            repository: repository fixture
            storage_format: storage format of database tables

        """
        version = "1.0.0"
        publish_db(build_dir, db, repository, version, storage_format)

        db = audb.load(db.name, version=version, verbose=False, full_path=False)
        assert_db_saved_to_dir(db, db.root, [storage_format, "pkl"])

    @pytest.mark.parametrize("pickle_tables", [True, False])
    def test_updated_database_save(
        self, build_dir, db, repository, storage_format, pickle_tables
    ):
        r"""Test correct files are stored to build dir after database update.

        Args:
            build_dir: build dir fixture
            db: database fixture
            repository: repository fixture
            storage_format: storage format of database tables
            pickle_tables: if ``True``,
                ``audb.load_to()`` should store tables
                as pickle files as well

        """
        # Publish first version
        version = "1.0.0"
        publish_db(build_dir, db, repository, version, storage_format)

        # Clear build dir to force audb.load_to() to load from backend
        audeer.rmdir(build_dir)
        db = audb.load_to(
            build_dir,
            db.name,
            version=version,
            pickle_tables=pickle_tables,
            verbose=False,
        )

        # Update database
        db = update_db(db)
        db.save(build_dir, storage_format=storage_format)
        if pickle_tables:
            expected_formats = [storage_format, "pkl"]
        else:
            expected_formats = [storage_format]
        assert_db_saved_to_dir(db, db.root, expected_formats)

    def test_updated_database_publish(
        self,
        build_dir: str,
        db: audformat.Database,
        repository: audb.Repository,
        storage_format: str,
    ):
        r"""Tests files in repository and dependency table after database update.

        Tables are stored differently in the repository
        for different storage formats,
        and their dependency table entry looks different as well.

        Args:
            build_dir: build dir fixture
            db: database fixture
            repository: repository fixture
            storage_format: database table storage format

        """
        # Publish first version
        previous_version = "1.0.0"
        publish_db(build_dir, db, repository, previous_version, storage_format)

        # Update database
        db = update_db(db)

        # Publish second version
        version = "1.1.0"
        deps = publish_db(build_dir, db, repository, version, storage_format)

        assert_db_published_to_repo(db, deps, repository, version, storage_format)
        assert_dependency_table(db, build_dir, deps, storage_format)

    def test_updated_database_load(
        self,
        build_dir: str,
        db: audformat.Database,
        repository: audb.Repository,
        storage_format: str,
    ):
        r"""Test correct files in cache after loading updated database.

        Args:
            build_dir: build dir fixture
            db: database fixture
            repository: repository fixture
            storage_format: storage format of database tables

        """
        # Publish first version
        previous_version = "1.0.0"
        publish_db(build_dir, db, repository, previous_version, storage_format)

        # Update database
        db = update_db(db)

        # Publish second version
        version = "1.1.0"
        publish_db(build_dir, db, repository, version, storage_format)

        # Load database to cache
        db = audb.load(db.name, version=version, verbose=False, full_path=False)
        assert_db_saved_to_dir(db, db.root, [storage_format, "pkl"])
        for table in list(db):
            assert "object" in db[table].df.columns


def test_publish_table_storage_format_both(db, build_dir, repository):
    r"""Test publishing of tables stored in CSV and PARQUET.

    When publishing tables,
    stored in CSV and PARQUET files at the same time,
    the CSV file should be ignored.

    Args:
        db: db fixture
        build_dir: build_dir fixture
        repository: repository fixture,
            providing a non-persistent repository
            on a file-system backend

    """
    # Save table as PARQUET and CSV at the same time
    db.save(build_dir, storage_format="parquet")
    db.save(build_dir, storage_format="csv", update_other_formats=False)

    # Check database build_dir looks as expected
    assert_db_saved_to_dir(db, build_dir, ["csv", "parquet"])

    # Publish database
    version = "1.0.0"
    deps = audb.publish(build_dir, version, repository)

    # Check files are published to repository
    assert_db_published_to_repo(db, deps, repository, version, "parquet")

    # Update only CSV table
    db = audb.load_to(build_dir, db.name, version="1.0.0", verbose=False)
    db_new = update_db(db)
    for table in list(db):
        db_new[table].save(
            audeer.path(build_dir, f"db.{table}"),
            storage_format="csv",
            update_other_formats=False,
        )
        # Remove PKL file to ensure CSV file is not newer
        os.remove(audeer.path(build_dir, f"db.{table}.pkl"))

    # Check database build_dir looks as expected
    assert_db_saved_to_dir(db, db.root, ["csv", "parquet"])

    # Publishing database from updated build_dir
    version = "1.1.0"
    deps = audb.publish(build_dir, version, repository, previous_version="1.0.0")

    # Check files are published to repository.
    # Only header and dependency table change,
    # as the PARQUET table stayed the same
    assert_db_published_to_repo(
        db,
        deps,
        repository,
        version,
        "parquet",
        table_changed=False,
    )

    # Load database to cache
    db = audb.load(db.name, version=version, verbose=False, full_path=False)
    assert_db_saved_to_dir(db, db.root, ["parquet", "pkl"])
    for table in list(db):
        assert "object" not in db[table].df.columns


def test_publish_table_storage_format_pkl(db, build_dir, repository):
    r"""Test publishing of tables stored in PKL.

    When publishing tables,
    stored only in PKL format,
    they should be converted to PARQUET automatically.

    Args:
        db: db fixture
        build_dir: build_dir fixture
        repository: repository fixture,
            providing a non-persistent repository
            on a file-system backend

    """
    # Save database with pickle tables
    db.save(build_dir, storage_format="pkl")
    assert_db_saved_to_dir(db, db.root, ["pkl"])

    # Publish database
    version = "1.0.0"
    deps = audb.publish(build_dir, version, repository)
    for table in list(db):
        assert f"db.{table}.parquet" in deps

    # Check files are published to repository
    assert_db_published_to_repo(db, deps, repository, version, "parquet")


def test_publish_table_parquet_without_hash(db, build_dir, repository):
    r"""Ensure publication of tables without hash metadata work.

    Args:
        db: db fixture
        build_dir: build_dir fixture
        repository: repository fixture

    """
    table_id = list(db)[0]
    table_file = audeer.path(build_dir, f"db.{table_id}.parquet")

    # Store parquet table without metadata entry
    db.save(build_dir, storage_format="parquet")
    table = parquet.read_table(table_file)
    metadata = table.schema.metadata.copy()
    del metadata[b"hash"]
    table = table.replace_schema_metadata(metadata)
    parquet.write_table(table, table_file, compression="snappy")

    # Check hash is not in metadata
    metadata = parquet.read_metadata(table_file)
    assert b"hash" not in metadata.metadata

    # Publish database
    version = "1.0.0"
    deps = audb.publish(build_dir, version, repository)
    assert f"db.{table_id}.parquet" in deps



import filecmp
import os
import re
import shutil

import numpy as np
import pandas as pd
import pytest

import audbackend
import audeer
import audformat.testing
import audiofile

import audb


DB_NAME = "test_publish"
LONG_PATH = "/".join(["audio"] * 50) + "/new.wav"


@pytest.fixture(
    scope="session",
    autouse=True,
)
def dbs(tmpdir_factory):
    r"""Store different versions of the same database.

    Returns:
        dictionary containing root folder for each version

    """
    # Collect single database paths
    # and return them in the end
    paths = {}

    # Version 0.1.0
    #
    # Folder without content
    version = "0.1.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = db_root

    # Version 1.0.0
    #
    # tables:
    #   - emotion
    #   - files
    # misc tables:
    #   - misc-in-scheme
    #   - misc-not-in-scheme
    # media:
    #   - audio/001.wav
    #   - audio/002.wav
    #   - audio/003.wav
    #   - audio/004.wav
    #   - audio/005.wav
    # attachment files:
    #   - extra/file.txt
    #   - extra/folder/file1.txt
    #   - extra/folder/file2.txt
    #   - extra/folder/sub-folder/file3.txt
    # schemes:
    #   - speaker
    #   - misc
    version = "1.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)
    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME
    db.schemes["scheme"] = audformat.Scheme(labels=["positive", "neutral", "negative"])
    audformat.testing.add_table(
        db,
        "emotion",
        audformat.define.IndexType.SEGMENTED,
        num_files=5,
        columns={"emotion": ("scheme", None)},
    )
    audformat.testing.add_misc_table(
        db,
        "misc-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    audformat.testing.add_misc_table(
        db,
        "misc-not-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    db.schemes["speaker"] = audformat.Scheme(labels=["adam", "11"])
    db.schemes["misc"] = audformat.Scheme(
        "int",
        labels="misc-in-scheme",
    )
    db["files"] = audformat.Table(db.files)
    db["files"]["speaker"] = audformat.Column(scheme_id="speaker")
    db["files"]["speaker"].set(
        ["adam", "adam", "adam", "11"],
        index=audformat.filewise_index(db.files[:4]),
    )
    db["files"]["misc"] = audformat.Column(scheme_id="misc")
    db["files"]["misc"].set(
        [0, 1, 1, 2],
        index=audformat.filewise_index(db.files[:4]),
    )
    audeer.mkdir(db_root, "extra/folder")
    audeer.touch(db_root, "extra/file.txt")
    audeer.touch(db_root, "extra/folder/file1.txt")
    audeer.touch(db_root, "extra/folder/file2.txt")
    audeer.mkdir(db_root, "extra/folder/sub-folder")
    audeer.touch(db_root, "extra/folder/sub-folder/file3.txt")
    # Create one file with different content to force different checksum
    file_with_different_content = audeer.path(
        db_root,
        "extra/folder/sub-folder/file3.txt",
    )
    with open(file_with_different_content, "w") as fp:
        fp.write("test")
    db.attachments["file"] = audformat.Attachment("extra/file.txt")
    db.attachments["folder"] = audformat.Attachment("extra/folder")
    db.save(
        db_root,
        storage_format=audformat.define.TableStorageFormat.PICKLE,
    )
    audformat.testing.create_audio_files(db)

    # Version 2.0.0
    #
    # Changes:
    #   * Added: new file with a path >260 characters
    #   * Added: 1 attachment file
    #   * Removed: 1 attachment file
    #
    # tables:
    #   - emotion
    #   - files
    # misc tables:
    #   - misc-in-scheme
    #   - misc-not-in-scheme
    # media:
    #   - audio/001.wav
    #   - audio/002.wav
    #   - audio/003.wav
    #   - audio/004.wav
    #   - audio/005.wav
    #   - audio/.../audio/new.wav  # >260 chars
    # attachment files:
    #   - extra/file.txt
    #   - extra/folder/file1.txt
    #   - extra/folder/file3.txt
    #   - extra/folder/sub-folder/file3.txt
    # schemes:
    #   - speaker
    #   - misc
    version = "2.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)
    shutil.copytree(
        audeer.path(paths["1.0.0"], "extra"),
        audeer.path(db_root, "extra"),
    )
    os.remove(audeer.path(db_root, "extra/folder/file2.txt"))
    audeer.touch(db_root, "extra/folder/file3.txt")
    db["files"].extend_index(audformat.filewise_index(LONG_PATH), inplace=True)
    db.save(db_root)
    audformat.testing.create_audio_files(db)

    # Version 2.1.0
    #
    # Folder without content
    version = "2.1.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)

    # Version 3.0.0
    #
    # Changes:
    #   * Removed: 1 media files
    #   * Removed: all attachments
    #
    # tables:
    #   - emotion
    #   - files
    # misc tables:
    #   - misc-in-scheme
    #   - misc-not-in-scheme
    # media:
    #   - audio/002.wav
    #   - audio/003.wav
    #   - audio/004.wav
    #   - audio/005.wav
    #   - audio/.../audio/new.wav  # >260 chars
    # schemes:
    #   - speaker
    #   - misc
    version = "3.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)
    remove_file = "audio/001.wav"
    db.drop_files(remove_file)
    del db.attachments["file"]
    del db.attachments["folder"]
    db.save(db_root)
    audformat.testing.create_audio_files(db)

    # Version 4.0.0
    #
    # Changes:
    #   * Changed: store without media files
    #
    # tables:
    #   - emotion
    #   - files
    # misc tables:
    #   - misc-in-scheme
    #   - misc-not-in-scheme
    # media:
    #   - audio/002.wav
    #   - audio/003.wav
    #   - audio/004.wav
    #   - audio/005.wav
    #   - audio/.../audio/new.wav  # >260 chars
    # schemes:
    #   - speaker
    #   - misc
    version = "4.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)
    db.save(db_root)

    # Version 5.0.0
    #
    # Changes:
    #   * Added: 20 `file{n}.wav` media files in metadata
    #
    # tables:
    #   - emotion
    #   - files
    # misc tables:
    #   - misc-in-scheme
    #   - misc-not-in-scheme
    # media:
    #   - audio/002.wav
    #   - audio/003.wav
    #   - audio/004.wav
    #   - audio/005.wav
    #   - audio/.../audio/new.wav  # >260 chars
    #   - file0.wav
    #   - ...
    #   - file19.wav
    # schemes:
    #   - speaker
    #   - misc
    version = "5.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)
    db["files"] = db["files"].extend_index(
        audformat.filewise_index([f"file{n}.wav" for n in range(20)])
    )
    assert len(db.files) > 20
    db.save(db_root)

    # Version 6.0.0
    #
    # Changes:
    #   * Added: 'scheme' scheme
    #   * Changed: include media files (not metadata only database)
    #   * Changed: make database non-portable
    #   * Removed: media file with >260 chars
    #   * Removed: 20 file{n}.wav media files
    #
    # tables:
    #   - emotion
    #   - files
    # misc tables:
    #   - misc-in-scheme
    #   - misc-not-in-scheme
    # media:
    #   - audio/001.wav
    #   - audio/002.wav
    #   - audio/003.wav
    #   - audio/004.wav
    #   - audio/005.wav
    # schemes:
    #   - scheme
    #   - speaker
    #   - misc
    version = "6.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)
    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME
    db.schemes["scheme"] = audformat.Scheme(labels=["positive", "neutral", "negative"])
    audformat.testing.add_table(
        db,
        "emotion",
        audformat.define.IndexType.SEGMENTED,
        num_files=5,
        columns={"emotion": ("scheme", None)},
    )
    audformat.testing.add_misc_table(
        db,
        "misc-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    audformat.testing.add_misc_table(
        db,
        "misc-not-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    db.schemes["speaker"] = audformat.Scheme(labels=["adam", "11"])
    db.schemes["misc"] = audformat.Scheme(
        "int",
        labels="misc-in-scheme",
    )
    db["files"] = audformat.Table(db.files)
    db["files"]["speaker"] = audformat.Column(scheme_id="speaker")
    db["files"]["speaker"].set(
        ["adam", "adam", "adam", "11"],
        index=audformat.filewise_index(db.files[:4]),
    )
    db["files"]["misc"] = audformat.Column(scheme_id="misc")
    db["files"]["misc"].set(
        [0, 1, 1, 2],
        index=audformat.filewise_index(db.files[:4]),
    )
    db.save(db_root)
    audformat.testing.create_audio_files(db)
    db.map_files(lambda x: os.path.join(db.root, x))  # make paths absolute
    db.save(db_root)

    return paths


@pytest.mark.parametrize(
    "name",
    ["?", "!", ","],
)
def test_invalid_archives(dbs, persistent_repository, name):
    archives = {"audio/001.wav": name}
    with pytest.raises(ValueError):
        audb.publish(
            dbs["1.0.0"],
            "1.0.1",
            persistent_repository,
            archives=archives,
            num_workers=pytest.NUM_WORKERS,
            verbose=False,
        )


@pytest.mark.parametrize(
    "version",
    [
        "1.0.0",
        pytest.param(
            "1.0.0",
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        "2.0.0",
        "3.0.0",
        pytest.param(
            "4.0.0",
            marks=pytest.mark.xfail(
                raises=RuntimeError,
                reason="Files missing (fewer than 20)",
            ),
        ),
        pytest.param(
            "5.0.0",
            marks=pytest.mark.xfail(
                raises=RuntimeError,
                reason="Files missing (more than 20)",
            ),
        ),
        pytest.param(
            "6.0.0",
            marks=pytest.mark.xfail(
                raises=RuntimeError,
                reason="Database not portable",
            ),
        ),
    ],
)
def test_publish(tmpdir, dbs, persistent_repository, version):
    db = audformat.Database.load(dbs[version])

    if not audb.versions(DB_NAME):
        with pytest.raises(RuntimeError):
            audb.latest_version(DB_NAME)

    # Copy database folder to build folder
    # to avoid storing dependency table files
    # inside the database folders
    build_dir = audeer.path(tmpdir, "build")
    shutil.copytree(dbs[version], build_dir)
    archives = db["files"]["speaker"].get().dropna().to_dict()
    deps = audb.publish(
        build_dir,
        version,
        persistent_repository,
        archives=archives,
        previous_version=None,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    backend_interface = audb.core.utils.lookup_backend(DB_NAME, version)
    number_of_media_files_in_custom_archives = len(set(archives.keys()))
    number_of_custom_archives = len(set(archives.values()))
    number_of_media_files = len(deps.media)
    number_of_media_archives = len({deps.archive(file) for file in deps.media})
    assert (number_of_media_files_in_custom_archives - number_of_custom_archives) == (
        number_of_media_files - number_of_media_archives
    )
    # Check if media files are sorted.
    # This does mean that media files are
    # always sorted by alphabetical order
    # but only in this specific test case.
    # Here we're testing for determinism rather
    # than ordering
    assert deps.media == sorted(deps.media)

    for archive in set(archives.values()):
        assert archive in deps.archives

    # Check checksums of attachment files
    for path in deps.attachments:
        expected_checksum = audeer.md5(audeer.path(db.root, path))
        assert deps.checksum(path) == expected_checksum

    db = audb.load(
        DB_NAME,
        version=version,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    assert db.name == DB_NAME

    versions = audb.versions(DB_NAME)
    latest_version = audb.latest_version(DB_NAME)

    assert version in versions
    assert latest_version == versions[-1]

    df = audb.available(only_latest=False)
    assert DB_NAME in df.index
    assert set(df[df.index == DB_NAME]["version"]) == set(versions)

    df = audb.available(only_latest=True)
    assert DB_NAME in df.index
    assert df[df.index == DB_NAME]["version"].iat[0] == latest_version

    for file in db.files:
        name = archives[file] if file in archives else file
        file_path = backend_interface.join("/", db.name, "media", name)
        backend_interface.exists(file_path, version)
        path = os.path.join(dbs[version], file)
        assert deps.checksum(file) == audeer.md5(path)
        if deps.format(file) in [
            audb.core.define.Format.WAV,
            audb.core.define.Format.FLAC,
        ]:
            assert deps.bit_depth(file) == audiofile.bit_depth(path)
            assert deps.channels(file) == audiofile.channels(path)
            assert deps.duration(file) == audiofile.duration(path)
            assert deps.sampling_rate(file) == audiofile.sampling_rate(path)


def test_publish_attachment(tmpdir, repository):
    # Create database (path does not need to exist)
    file_path = "attachments/file.txt"
    folder_path = "attachments/folder"
    db = audformat.Database("db-with-attachments")
    db.attachments["file"] = audformat.Attachment(
        file_path,
        description="Attached file",
        meta={"mime": "text"},
    )
    db.attachments["folder"] = audformat.Attachment(
        folder_path,
        description="Attached folder",
        meta={"mime": "inode/directory"},
    )

    assert list(db.attachments) == ["file", "folder"]
    assert db.attachments["file"].path == file_path
    assert db.attachments["file"].description == "Attached file"
    assert db.attachments["file"].meta == {"mime": "text"}
    assert db.attachments["folder"].path == folder_path
    assert db.attachments["folder"].description == "Attached folder"
    assert db.attachments["folder"].meta == {"mime": "inode/directory"}

    db_path = audeer.path(tmpdir, "db")
    audeer.mkdir(db_path)
    db.save(db_path)

    # Publish database, path needs to exist
    error_msg = (
        f"The provided path '{file_path}' " f"of attachment 'file' " "does not exist."
    )
    with pytest.raises(FileNotFoundError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)

    # Publish database, path is not allowed to be a symlink
    audeer.mkdir(db_path, folder_path)
    os.symlink(
        audeer.path(db_path, folder_path),
        audeer.path(db_path, file_path),
    )
    error_msg = (
        f"The provided path '{file_path}' "
        f"of attachment 'file' "
        "must not be a symlink."
    )
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)

    os.remove(os.path.join(db_path, file_path))
    audeer.touch(db_path, file_path)
    db.save(db_path)

    # File exist now, folder is empty
    assert db.attachments["file"].files == [file_path]
    assert db.attachments["folder"].files == []
    error_msg = (
        "An attached folder must "
        "contain at least one file. "
        "But attachment 'folder' "
        "points to an empty folder."
    )
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)

    # Add empty sub-folder
    subfolder_path = f"{folder_path}/sub-folder"
    audeer.mkdir(db_path, subfolder_path)
    assert db.attachments["folder"].files == []
    error_msg = (
        "An attachment must not "
        "contain empty sub-folders. "
        "But attachment 'folder' "
        "contains the empty sub-folder 'sub-folder'."
    )
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)

    # Add file to folder, sub-folder still empty
    file2_path = f"{folder_path}/file.txt"
    audeer.touch(db_path, file2_path)
    assert db.attachments["file"].files == [file_path]
    assert db.attachments["folder"].files == [file2_path]
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)

    # Add file to sub-folder
    file3_path = f"{subfolder_path}/file.txt"
    audeer.touch(db_path, file3_path)
    assert db.attachments["file"].files == [file_path]
    assert db.attachments["folder"].files == [file2_path, file3_path]

    # Publish and load database
    audb.publish(db_path, "1.0.0", repository)
    db = audb.load(db.name, version="1.0.0", verbose=False)
    assert list(db.attachments) == ["file", "folder"]
    assert db.attachments["file"].files == [file_path]
    assert db.attachments["folder"].files == [file2_path, file3_path]
    assert db.attachments["file"].path == file_path
    assert db.attachments["file"].description == "Attached file"
    assert db.attachments["file"].meta == {"mime": "text"}
    assert db.attachments["folder"].path == folder_path
    assert db.attachments["folder"].description == "Attached folder"
    assert db.attachments["folder"].meta == {"mime": "inode/directory"}


def test_publish_change_archive(tmpdir, dbs, repository):
    """Test removing file from archive.

    When a file is removed,
    which is stored in an archive
    together with other files,
    the archive needs to be updated.

    As described in
    https://github.com/audeering/audb/issues/377
    this was failing
    when the archive was not added
    in ``previous_version``,
    but an earlier version.

    """
    # Store two files in an archive in version 1.0.0
    build_dir = audeer.path(tmpdir, "build")
    shutil.copytree(dbs["1.0.0"], build_dir)
    archives = {
        "audio/001.wav": "common_archive",
        "audio/002.wav": "common_archive",
    }
    audb.publish(
        build_dir,
        "1.0.0",
        repository,
        archives=archives,
        previous_version=None,
        verbose=False,
    )
    # Remove a file stored in another archive for version 2.0.0
    os.remove(audeer.path(build_dir, "audio/003.wav"))
    audb.publish(
        build_dir,
        "2.0.0",
        repository,
        archives=archives,
        previous_version="1.0.0",
        verbose=False,
    )
    # Remove file `audio/001.wav`
    # to require archive update
    # of archive added in version "1.0.0"
    audeer.rmdir(build_dir)
    db = audb.load_to(
        build_dir,
        DB_NAME,
        version="2.0.0",
        only_metadata=True,
    )
    db.drop_files(["audio/001.wav"])
    db.save(build_dir)
    audb.publish(
        build_dir,
        "3.0.0",
        repository,
        previous_version="2.0.0",
        verbose=False,
    )


@pytest.mark.parametrize(
    "version1, version2, media_difference, attachment_difference",
    [
        (
            "1.0.0",
            "1.0.0",
            [],
            [],
        ),
        (
            "1.0.0",
            "2.0.0",
            [],
            [os.path.join("extra", "folder", "file2.txt")],
        ),
        (
            "2.0.0",
            "1.0.0",
            [LONG_PATH],
            [os.path.join("extra", "folder", "file3.txt")],
        ),
        (
            "2.0.0",
            "3.0.0",
            ["audio/001.wav"],
            [
                os.path.join("extra", "file.txt"),
                os.path.join("extra", "folder", "file1.txt"),
                os.path.join("extra", "folder", "file3.txt"),
                os.path.join("extra", "folder", "sub-folder", "file3.txt"),
            ],
        ),
        (
            "3.0.0",
            "2.0.0",
            [],
            [],
        ),
        (
            "1.0.0",
            "3.0.0",
            ["audio/001.wav"],
            [
                os.path.join("extra", "file.txt"),
                os.path.join("extra", "folder", "file1.txt"),
                os.path.join("extra", "folder", "file2.txt"),
                os.path.join("extra", "folder", "sub-folder", "file3.txt"),
            ],
        ),
        (
            "3.0.0",
            "1.0.0",
            [LONG_PATH],
            [],
        ),
    ],
)
def test_publish_changed_db(
    dbs,
    version1,
    version2,
    media_difference,
    attachment_difference,
):
    depend1 = audb.dependencies(DB_NAME, version=version1)
    depend2 = audb.dependencies(DB_NAME, version=version2)

    media1 = set(sorted(depend1.media))
    media2 = set(sorted(depend2.media))
    assert media1 - media2 == set(media_difference)

    attachment1 = []
    for path in depend1.attachments:
        root = dbs[version1]
        files = audeer.list_file_names(
            audeer.path(root, path),
            recursive=True,
            hidden=True,
        )
        files = [file[len(root) + 1 :] for file in files]
        attachment1.extend(files)

    attachment2 = []
    for path in depend2.attachments:
        root = dbs[version2]
        files = audeer.list_file_names(
            audeer.path(root, path),
            recursive=True,
            hidden=True,
        )
        files = [file[len(root) + 1 :] for file in files]
        attachment2.extend(files)

    assert set(attachment1) - set(attachment2) == set(attachment_difference)


@pytest.mark.parametrize(
    "version, previous_version, error_type, error_msg",
    [
        (
            "1.0.0",
            None,
            RuntimeError,
            ("A version '1.0.0' already exists for database " f"'{DB_NAME}'."),
        ),
        (
            "4.0.0",
            None,
            RuntimeError,
            (
                "The following 5 files are referenced in tables "
                "that cannot be found on disk "
                "and are not yet part of the database: "
                "['audio/002.wav', 'audio/003.wav', "
                "'audio/004.wav', 'audio/005.wav', "
                f"'{LONG_PATH}']."
            ),
        ),
        (
            "5.0.0",
            None,
            RuntimeError,
            (
                "The following 25 files are referenced in tables "
                "that cannot be found on disk "
                "and are not yet part of the database: "
                "['audio/002.wav', 'audio/003.wav', "
                "'audio/004.wav', 'audio/005.wav', "
                f"'{LONG_PATH}', "
                "'file0.wav', 'file1.wav', 'file10.wav', 'file11.wav', "
                "'file12.wav', 'file13.wav', 'file14.wav', 'file15.wav', "
                "'file16.wav', 'file17.wav', 'file18.wav', 'file19.wav', "
                "'file2.wav', 'file3.wav', 'file4.wav', ...]."
            ),
        ),
        (
            "5.0.0",
            "1.0.0",
            RuntimeError,
            (
                "The following 21 files are referenced in tables "
                "that cannot be found on disk "
                "and are not yet part of the database: "
                f"['{LONG_PATH}', "
                "'file0.wav', 'file1.wav', 'file10.wav', 'file11.wav', "
                "'file12.wav', 'file13.wav', 'file14.wav', 'file15.wav', "
                "'file16.wav', 'file17.wav', 'file18.wav', 'file19.wav', "
                "'file2.wav', 'file3.wav', 'file4.wav', 'file5.wav', "
                "'file6.wav', 'file7.wav', 'file8.wav', ...]."
            ),
        ),
        (
            "0.1.0",
            "1.0.0",
            ValueError,
            (
                "'previous_version' needs to be smaller than 'version', "
                "but yours is 1.0.0 >= 0.1.0."
            ),
        ),
        (
            "0.1.0",
            "0.1.0",
            ValueError,
            (
                "'previous_version' needs to be smaller than 'version', "
                "but yours is 0.1.0 >= 0.1.0."
            ),
        ),
    ],
)
def test_publish_error_messages(
    dbs,
    persistent_repository,
    version,
    previous_version,
    error_type,
    error_msg,
):
    with pytest.raises(error_type, match=re.escape(error_msg)):
        if previous_version and (
            audeer.StrictVersion(previous_version) < audeer.StrictVersion(version)
        ):
            deps = audb.dependencies(
                DB_NAME,
                version=previous_version,
            )
            path = os.path.join(
                dbs[version],
                audb.core.define.DEPENDENCY_FILE,
            )
            deps.save(path)
        audb.publish(
            dbs[version],
            version,
            persistent_repository,
            previous_version=previous_version,
            num_workers=pytest.NUM_WORKERS,
            verbose=False,
        )


def test_publish_error_allowed_chars(tmpdir, repository):
    # Table and attachment IDs are only allow
    # to contain chars that can be used as filenames
    # on the backends

    # Prepare database files
    db_path = audeer.mkdir(tmpdir, "db")
    audio_file = audeer.path(db_path, "f1.wav")
    attachment_file = audeer.path(db_path, "attachment.txt")
    signal = np.zeros((2, 1000))
    sampling_rate = 8000
    audiofile.write(audio_file, signal, sampling_rate)
    audeer.touch(attachment_file)

    # Database with not allowed table ID
    db = audformat.Database("db")
    index = audformat.filewise_index(os.path.basename(audio_file))
    db["table/"] = audformat.Table(index)
    db["table/"]["column"] = audformat.Column()
    db["table/"]["column"].set(["label"])
    db.save(db_path)
    error_msg = (
        "Table IDs must only contain chars from [A-Za-z0-9._-], "
        "which is not the case for table 'table/'."
    )
    with pytest.raises(RuntimeError, match=re.escape(error_msg)):
        audb.publish(db_path, "1.0.0", repository)

    # Database with not allowed attachment ID
    db = audformat.Database("db")
    index = audformat.filewise_index(os.path.basename(audio_file))
    db["table"] = audformat.Table(index)
    db["table"]["column"] = audformat.Column()
    db["table"]["column"].set(["label"])
    db.attachments["attachment?"] = audformat.Attachment(
        os.path.basename(attachment_file)
    )
    db.save(db_path)
    error_msg = (
        "Attachment IDs must only contain chars from [A-Za-z0-9._-], "
        "which is not the case for attachment 'attachment?'."
    )
    with pytest.raises(RuntimeError, match=re.escape(error_msg)):
        audb.publish(db_path, "1.0.0", repository)

    # Ensure column and scheme IDs are not affected
    db = audformat.Database("db")
    index = audformat.filewise_index(os.path.basename(audio_file))
    db.schemes["scheme?"] = audformat.Scheme("str")
    db["table"] = audformat.Table(index)
    db["table"]["column?"] = audformat.Column(scheme_id="scheme?")
    db["table"]["column?"].set(["label"])
    db.save(db_path)
    audb.publish(db_path, "1.0.0", repository)


def test_publish_error_changed_deps_file_type(tmpdir, repository):
    # As we allow for every possible filename for attachments
    # and store them in the dependency table
    # besides media and table files
    # there can be a naming clash between those entries.
    # See https://github.com/audeering/audb/pull/244#issuecomment-1412211131

    # media => attachment
    error_msg = (
        "An attachment must not overlap with media or tables. "
        "But attachment 'attachment' contains 'data/file.wav'."
    )
    db_name = "test_publish_error_changed_deps_file_type-1"
    db_path = audeer.mkdir(tmpdir, "db")
    data_path = audeer.mkdir(db_path, "data")
    signal = np.zeros((2, 1000))
    sampling_rate = 8000
    audiofile.write(audeer.path(data_path, "file.wav"), signal, sampling_rate)
    db = audformat.Database(db_name)
    db["table"] = audformat.Table(audformat.filewise_index("data/file.wav"))
    db.attachments["attachment"] = audformat.Attachment("data/file.wav")
    db.save(db_path)
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)
    audeer.rmdir(db_path)

    # table => attachment
    error_msg = (
        "An attachment must not overlap with media or tables. "
        "But attachment 'attachment' contains 'db.table.csv'."
    )
    db_name = "test_publish_error_changed_deps_file_type-2"
    db_path = audeer.mkdir(tmpdir, "db")
    data_path = audeer.mkdir(db_path, "data")
    signal = np.zeros((2, 1000))
    sampling_rate = 8000
    audiofile.write(audeer.path(data_path, "file.wav"), signal, sampling_rate)
    db = audformat.Database(db_name)
    db["table"] = audformat.Table(audformat.filewise_index("data/file.wav"))
    db.attachments["attachment"] = audformat.Attachment("db.table.csv")
    db.save(db_path)
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)
    audeer.rmdir(db_path)

    # attachment => media
    error_msg = (
        "An attachment must not overlap with media or tables. "
        "But attachment 'attachment' contains 'data/file2.wav'."
    )
    db_name = "test_publish_error_changed_deps_file_type-3"
    db_path = audeer.mkdir(tmpdir, "db")
    data_path = audeer.mkdir(db_path, "data")
    signal = np.zeros((2, 1000))
    sampling_rate = 8000
    audiofile.write(audeer.path(data_path, "file1.wav"), signal, sampling_rate)
    audiofile.write(audeer.path(data_path, "file2.wav"), signal, sampling_rate)
    db = audformat.Database(db_name)
    db["table"] = audformat.Table(audformat.filewise_index("data/file1.wav"))
    db.attachments["attachment"] = audformat.Attachment("data/file2.wav")
    db.save(db_path)
    audb.publish(db_path, "1.0.0", repository)
    audeer.rmdir(db_path)
    db = audb.load_to(db_path, db_name, version="1.0.0")
    db["table"] = audformat.Table(
        audformat.filewise_index(["data/file1.wav", "data/file2.wav"])
    )
    db.save(db_path)
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "2.0.0", repository)
    audeer.rmdir(db_path)

    # attachment => table
    error_msg = (
        "An attachment must not overlap with media or tables. "
        "But attachment 'attachment' contains 'db.table2.csv'."
    )
    db_name = "test_publish_error_changed_deps_file_type-4"
    db_path = audeer.mkdir(tmpdir, "db")
    data_path = audeer.mkdir(db_path, "data")
    signal = np.zeros((2, 1000))
    sampling_rate = 8000
    audiofile.write(audeer.path(data_path, "file.wav"), signal, sampling_rate)
    db = audformat.Database(db_name)
    db["table1"] = audformat.Table(audformat.filewise_index("data/file.wav"))
    db.attachments["attachment"] = audformat.Attachment("db.table2.csv")
    audeer.touch(db_path, "db.table2.csv")
    db.save(db_path)
    audb.publish(db_path, "1.0.0", repository)
    audeer.rmdir(db_path)
    db = audb.load_to(db_path, db_name, version="1.0.0")
    db["table2"] = audformat.Table(audformat.filewise_index("data/file.wav"))
    db.save(db_path)
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "2.0.0", repository)
    audeer.rmdir(db_path)


def test_publish_error_repository_does_not_exist(tmpdir, repository):
    db = audformat.Database("test")
    db.save(tmpdir)

    repository.name = "does-not-exist"
    with pytest.raises(audbackend.BackendError) as ex:
        audb.publish(tmpdir, "1.0.0", repository)
    assert "No such file or directory" in str(ex.value.exception)


@pytest.mark.parametrize(
    "file",
    [
        "file.Wav",
        "file.WAV",
        "file.1A",
    ],
)
def test_publish_error_uppercase_file_extension(tmpdir, repository, file):
    # Prepare files
    db_path = audeer.mkdir(tmpdir, "db")
    audeer.touch(db_path, file)
    # Prepare database
    db = audformat.Database("db")
    db["table"] = audformat.Table(audformat.filewise_index([file]))
    db["table"]["column"] = audformat.Column()
    db["table"]["column"].set(["label"])
    db.save(db_path)
    # Fail as we include file with uppercase letter
    error_msg = (
        "The file extension of a media file must be lowercase, "
        f"but '{file}' includes at least one uppercase letter."
    )
    with pytest.raises(RuntimeError, match=error_msg):
        audb.publish(db_path, "1.0.0", repository)


def test_publish_error_version(tmpdir, repository):
    # Only versions supported by audeer.StrictVersion
    # are allowed

    # Create simple database
    db_path = audeer.mkdir(tmpdir, "db")
    audio_file = audeer.path(db_path, "f1.wav")
    signal = np.zeros((2, 1000))
    sampling_rate = 8000
    audiofile.write(audio_file, signal, sampling_rate)
    db = audformat.Database("db")
    index = audformat.filewise_index(os.path.basename(audio_file))
    db["table"] = audformat.Table(index)
    db["table"]["column"] = audformat.Column()
    db["table"]["column"].set(["label"])
    db.save(db_path)

    error_msg = "invalid version number '1.0.0?'"
    with pytest.raises(ValueError, match=re.escape(error_msg)):
        audb.publish(db_path, "1.0.0?", repository)

    # Publish to check previous_version afterwards
    audb.publish(db_path, "1.0.0", repository)

    # Update database
    db["table"]["column"].set(["different-label"])
    db.save(db_path)

    error_msg = "invalid version number '1.0.0?'"
    with pytest.raises(ValueError, match=re.escape(error_msg)):
        audb.publish(db_path, "2.0.0", repository, previous_version="1.0.0?")


def test_publish_text_media_files(tmpdir, dbs, repository, storage_format):
    r"""Test publishing databases containing text files as media files."""
    # Create a database, containing text media file
    build_dir = audeer.path(tmpdir, "./build")
    audeer.mkdir(build_dir)
    data_dir = audeer.mkdir(build_dir, "data")
    with open(audeer.path(data_dir, "file1.txt"), "w") as file:
        file.write("Text written by a person.\n")
    name = "text-db"
    db = audformat.Database(name)
    db.schemes["speaker"] = audformat.Scheme("str")
    index = audformat.filewise_index(["data/file1.txt"])
    db["files"] = audformat.Table(index)
    db["files"]["speaker"] = audformat.Column(scheme_id="speaker")
    db["files"]["speaker"].set(["adam"])
    db.save(build_dir, storage_format=storage_format)

    # Publish database, containing text media file
    version = "1.0.0"
    deps = audb.publish(build_dir, version, repository)

    assert deps.tables == [f"db.files.{storage_format}"]
    file = "data/file1.txt"
    assert deps.media == [file]
    assert deps.bit_depth(file) == 0
    assert deps.channels(file) == 0
    assert deps.duration(file) == 0.0
    assert deps.format(file) == "txt"
    assert deps.sampling_rate(file) == 0

    db = audb.load(name, version=version, verbose=False, full_path=False)
    assert db.files == [file]
    assert list(db) == ["files"]
    assert os.path.exists(audeer.path(db.root, file))

    error_msg = f"Media file '{file}' does not support requesting a flavor."
    with pytest.raises(RuntimeError, match=error_msg):
        db = audb.load(name, version=version, channels=[0], verbose=False)

    # Publish database, containing text and media files
    audeer.rmdir(build_dir)
    shutil.copytree(dbs["1.0.0"], build_dir)  # start with db containing audio files
    db = audformat.Database.load(build_dir)
    speaker = db["files"]["speaker"].get()
    files = list(db.files)
    tables = list(db)
    data_dir = audeer.mkdir(build_dir, "data")
    with open(audeer.path(data_dir, "file1.txt"), "w") as file:
        file.write("Text written by a person.\n")
    index = audformat.filewise_index(["data/file1.txt"])
    db["files"].extend_index(index, inplace=True)
    db["files"]["speaker"] = audformat.Column(scheme_id="speaker")
    db["files"]["speaker"].set(list(speaker.values) + ["adam"])
    db.name = name
    db.save(build_dir)

    # Publish database, containing text media file
    version = "2.0.0"
    deps = audb.publish(build_dir, version, repository, previous_version=None)

    assert deps.table_ids == tables
    file = "data/file1.txt"
    assert deps.media == files + [file]
    assert deps.bit_depth(file) == 0
    assert deps.channels(file) == 0
    assert deps.duration(file) == 0.0
    assert deps.format(file) == "txt"
    assert deps.sampling_rate(file) == 0

    db = audb.load(name, version=version, verbose=False, full_path=False)
    assert db.files == files + [file]
    assert list(db) == tables
    assert os.path.exists(audeer.path(db.root, file))

    error_msg = f"Media file '{file}' does not support requesting a flavor."
    with pytest.raises(RuntimeError, match=error_msg):
        db = audb.load(name, version=version, channels=[0], verbose=False)


def test_update_database(dbs, persistent_repository):
    version = "2.1.0"
    start_version = "2.0.0"

    audb.load_to(
        dbs[version],
        DB_NAME,
        version=start_version,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )

    # == Fail with missing dependency file
    previous_version = start_version
    dep_file = os.path.join(
        dbs[version],
        audb.core.define.DEPENDENCY_FILE,
    )
    os.remove(dep_file)
    error_msg = (
        f"You want to depend on '{previous_version}' "
        f"of {DB_NAME}, "
        f"but you don't have a '{audb.core.define.DEPENDENCY_FILE}' "
        f"file present "
        f"in {dbs[version]}. "
        f"Did you forgot to call "
        f"'audb.load_to({dbs[version]}, {DB_NAME}, "
        f"version={previous_version}?"
    )
    with pytest.raises(RuntimeError, match=re.escape(error_msg)):
        audb.publish(
            dbs[version],
            version,
            persistent_repository,
            previous_version=previous_version,
            num_workers=pytest.NUM_WORKERS,
            verbose=False,
        )

    # Reload data to restore dependency file
    shutil.rmtree(dbs[version])
    db = audb.load_to(
        dbs[version],
        DB_NAME,
        version=start_version,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    # Remove one media file and all attachments as in version 3.0.0
    remove_file = "audio/001.wav"
    remove_path = os.path.join(dbs[version], remove_file)
    os.remove(remove_path)
    del db.attachments["file"]
    del db.attachments["folder"]
    db.drop_files(remove_file)
    db.save(dbs[version])

    # == Fail as 2.0.0 is not the latest version
    previous_version = "latest"
    error_msg = (
        f"You want to depend on '{audb.latest_version(DB_NAME)}' "
        f"of {DB_NAME}, "
        f"but the dependency file "
        f"'{audb.core.define.DEPENDENCY_FILE}' "
        f"in {dbs[version]} "
        f"does not match the dependency file "
        f"for the requested version in the repository. "
        f"Did you forgot to call "
        f"'audb.load_to({dbs[version]}, {DB_NAME}, "
        f"version='{audb.latest_version(DB_NAME)}') "
        f"or modified the file manually?"
    )
    with pytest.raises(RuntimeError, match=re.escape(error_msg)):
        audb.publish(
            dbs[version],
            version,
            persistent_repository,
            previous_version=previous_version,
            num_workers=pytest.NUM_WORKERS,
            verbose=False,
        )

    # == Fail as we require a previous version
    previous_version = None
    error_msg = (
        f"You did not set a dependency to a previous version, "
        f"but you have a '{audb.core.define.DEPENDENCY_FILE}' file present "
        f"in {dbs[version]}."
    )
    with pytest.raises(RuntimeError, match=re.escape(error_msg)):
        audb.publish(
            dbs[version],
            version,
            persistent_repository,
            previous_version=previous_version,
            num_workers=pytest.NUM_WORKERS,
            verbose=False,
        )

    previous_version = start_version
    deps = audb.publish(
        dbs[version],
        version,
        persistent_repository,
        previous_version=previous_version,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )

    # Check that dependencies include previous and actual version only
    versions = audeer.sort_versions([deps.version(f) for f in deps.files])
    assert versions[-1] == version
    assert versions[0] == previous_version

    # Check that there is no difference in the database
    # if published from scratch or from previous version
    db1 = audb.load(
        DB_NAME,
        version=version,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    db2 = audb.load(
        DB_NAME,
        version="3.0.0",
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    db1.meta["audb"] = {}
    db2.meta["audb"] = {}
    assert db1 == db2


def test_update_database_without_media(tmpdir, persistent_repository):
    build_root = tmpdir
    previous_version = "1.0.0"
    version = "1.1.0"

    # attachments
    new_attachment = "new-atatchment"
    # attachment files
    # (all for attachment ID `folder`)
    new_attachment_files = [
        "extra/folder/file4.txt",
    ]
    altered_attachment_files = [
        "extra/folder/sub-folder/file3.txt",
    ]
    removed_attachment_files = [
        "extra/folder/file1.txt",
        "extra/folder/file2.txt",
    ]

    # tables
    new_table = "new-table"

    # media
    new_files = [
        "new/001.wav",
        "new/002.wav",
    ]
    alter_files = [
        "audio/001.wav",  # same archive as 'audio/00[2,3].wav'
        "audio/005.wav",
    ]
    rem_files = [
        "new/003.wav",  # same archive as 'audio/00[1,2].wav'
    ]

    # load without media
    db = audb.load_to(
        build_root,
        DB_NAME,
        only_metadata=True,
        version=previous_version,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    for file in db.files:
        assert not os.path.exists(audeer.path(build_root, file))

    # add changes to build folder
    # and call again load_to()
    # to revert them

    os.remove(audeer.path(build_root, "db.emotion.parquet"))
    db = audb.load_to(
        build_root,
        DB_NAME,
        only_metadata=True,
        version=previous_version,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    assert os.path.exists(audeer.path(build_root, "db.emotion.parquet"))

    # update and save database

    # remove media files
    for file in rem_files:
        db.drop_files(file)

    # create and alter media files
    sampling_rate = 16000
    signal = np.ones((1, sampling_rate), np.float32)
    for file in new_files + alter_files:
        path = audeer.path(build_root, file)
        audeer.mkdir(os.path.dirname(path))
        audiofile.write(path, signal, sampling_rate)

    # add new table
    db[new_table] = audformat.Table(audformat.filewise_index(new_files))

    # Add, alter, and remove attachment files for `folder` attachment.
    # The removal is done automatically as we alter
    # and add a file in the same folder
    # as the to be removed attachment files are located.
    for attachment_file in new_attachment_files:
        path = audeer.path(build_root, attachment_file)
        audeer.mkdir(os.path.dirname(path))
        audeer.touch(path)
    for attachment_file in altered_attachment_files:
        path = audeer.path(build_root, attachment_file)
        audeer.mkdir(os.path.dirname(path))
        with open(path, "w") as fp:
            fp.write("abc")
    for attachment_file in removed_attachment_files:
        path = audeer.path(build_root, attachment_file)
        assert not os.path.exists(path)

    # add new attachment
    audeer.mkdir(build_root, "extra")
    audeer.touch(build_root, "extra/new.txt")
    db.attachments[new_attachment] = audformat.Attachment("extra/new.txt")

    db.save(build_root)

    # publish database
    audb.publish(
        build_root,
        version,
        persistent_repository,
        previous_version=previous_version,
        verbose=False,
    )

    # check if media files missing from the same archive
    # were downloaded during publish
    assert os.path.exists(audeer.path(build_root, "audio/002.wav"))

    # load new version and check media and attachments
    db_load = audb.load(
        DB_NAME,
        version=version,
        verbose=False,
    )
    for file in db.files:
        assert os.path.exists(audeer.path(db_load.root, file))
    for file in rem_files:
        assert not os.path.exists(audeer.path(db_load.root, file))
    for file in new_files + alter_files:
        assert filecmp.cmp(
            audeer.path(build_root, file),
            audeer.path(db_load.root, file),
        )
    assert new_attachment in list(db_load.attachments)
    assert db.attachments == db_load.attachments
    for attachment in db_load.attachments:
        path = audeer.path(db_load.root, db_load.attachments[attachment].path)
        assert os.path.exists(path)
        for attachment_file in db_load.attachments[attachment].files:
            assert os.path.exists(audeer.path(db_load.root, attachment_file))
    for attachment_file in removed_attachment_files:
        assert not os.path.exists(audeer.path(db_load.root, attachment_file))
    for attachment_file in new_attachment_files + altered_attachment_files:
        assert os.path.exists(audeer.path(db_load.root, attachment_file))
        assert filecmp.cmp(
            audeer.path(build_root, attachment_file),
            audeer.path(db_load.root, attachment_file),
        )



import pandas as pd
import pytest

import audformat.testing

import audb


DB_NAME = "test_filter"


@pytest.fixture(
    scope="module",
    autouse=True,
)
def db(tmpdir_factory, persistent_repository):
    r"""Publish a single database."""
    version = "1.0.0"
    db_root = tmpdir_factory.mktemp(version)

    # create db

    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME
    db.schemes["scheme"] = audformat.Scheme(labels=["some", "test", "labels"])
    audformat.testing.add_misc_table(
        db,
        "misc-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    audformat.testing.add_misc_table(
        db,
        "misc-not-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    db.schemes["misc"] = audformat.Scheme(
        "int",
        labels="misc-in-scheme",
    )
    audformat.testing.add_table(
        db,
        "test",
        audformat.define.IndexType.SEGMENTED,
        columns={"label": ("scheme", None)},
        num_files=[0, 1],
    )
    audformat.testing.add_table(
        db,
        "dev",
        audformat.define.IndexType.SEGMENTED,
        columns={"label": ("misc", None)},
        num_files=[10, 11],
    )
    audformat.testing.add_table(
        db,
        "train",
        audformat.define.IndexType.SEGMENTED,
        columns={"label": ("scheme", None)},
        num_files=[20, 21],
    )
    # Add nested folder structure to ensure not all files in an archive
    # are stored in the same folder
    mapping = {
        "audio/020.wav": "audio/1/020.wav",
        "audio/021.wav": "audio/2/021.wav",
    }
    files = db["train"].df.index.get_level_values("file")
    starts = db["train"].df.index.get_level_values("start")
    ends = db["train"].df.index.get_level_values("end")
    db["train"].df.index = audformat.segmented_index(
        files=[mapping[f] for f in files],
        starts=starts,
        ends=ends,
    )
    db.save(db_root)
    audformat.testing.create_audio_files(db)

    # publish db

    archives = {}
    for table in db.tables:
        archives.update({file: table for file in db[table].files})
    audb.publish(
        db_root,
        version,
        persistent_repository,
        archives=archives,
        verbose=False,
    )


@pytest.mark.parametrize(
    "media, format, expected_files",
    [
        (
            None,
            None,
            [
                "audio/000.wav",
                "audio/001.wav",
                "audio/010.wav",
                "audio/011.wav",
                "audio/1/020.wav",
                "audio/2/021.wav",
            ],
        ),
        (
            [],
            None,
            [],
        ),
        (
            "",
            None,
            [],
        ),
        (
            ["audio/000.wav", "audio/001.wav"],
            None,
            ["audio/000.wav", "audio/001.wav"],
        ),
        (
            ["audio/000.wav", "audio/001.wav"],
            "flac",
            ["audio/000.flac", "audio/001.flac"],
        ),
        (
            r".*0\.wav",
            None,
            ["audio/000.wav", "audio/010.wav", "audio/1/020.wav"],
        ),
        pytest.param(
            "non-existing",
            None,
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            ["non-existing"],
            None,
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_media(media, format, expected_files):
    db = audb.load(
        DB_NAME,
        media=media,
        format=format,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    assert list(db.files) == expected_files
    assert list(db.tables) == ["dev", "test", "train"]


@pytest.mark.parametrize(
    "tables, format, expected_tables, expected_files",
    [
        (
            None,
            None,
            ["dev", "misc-in-scheme", "misc-not-in-scheme", "test", "train"],
            [
                "audio/000.wav",
                "audio/001.wav",
                "audio/010.wav",
                "audio/011.wav",
                "audio/1/020.wav",
                "audio/2/021.wav",
            ],
        ),
        (
            "test",
            None,
            ["misc-in-scheme", "test"],
            ["audio/000.wav", "audio/001.wav"],
        ),
        (
            "^t.*",
            None,
            ["misc-in-scheme", "test", "train"],
            ["audio/000.wav", "audio/001.wav", "audio/1/020.wav", "audio/2/021.wav"],
        ),
        (
            ["dev", "train", "misc-not-in-scheme"],
            None,
            ["dev", "misc-in-scheme", "misc-not-in-scheme", "train"],
            ["audio/010.wav", "audio/011.wav", "audio/1/020.wav", "audio/2/021.wav"],
        ),
        (
            ["dev", "train"],
            "flac",
            ["dev", "misc-in-scheme", "train"],
            [
                "audio/010.flac",
                "audio/011.flac",
                "audio/1/020.flac",
                "audio/2/021.flac",
            ],
        ),
        (
            [],
            None,
            ["misc-in-scheme"],
            [],
        ),
        (
            "",
            None,
            ["misc-in-scheme"],
            [],
        ),
        pytest.param(
            "non-existing",
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            ["non-existing"],
            None,
            None,
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_tables(tables, format, expected_tables, expected_files):
    db = audb.load(
        DB_NAME,
        tables=tables,
        format=format,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    assert list(db) == expected_tables
    assert list(db.files) == expected_files



import os

import pytest

import audeer
import audformat.testing

import audb


DB_NAMES = [
    "test_cache-0",
    "test_cache-1",
]


@pytest.fixture(
    scope="module",
    autouse=True,
)
def dbs(tmpdir_factory, persistent_repository):
    # create dbs

    for name in DB_NAMES:
        db_root = tmpdir_factory.mktemp(name)
        db = audformat.testing.create_db(minimal=True)
        db.name = name
        db["files"] = audformat.Table(audformat.filewise_index(["f1.wav"]))

        db.save(db_root)
        audformat.testing.create_audio_files(db)
        audb.publish(
            db_root,
            "1.0.0",
            persistent_repository,
            verbose=False,
        )


@pytest.mark.parametrize("shared", [False, True])
def test_cache_root(cache, shared_cache, shared):
    if shared:
        assert audb.default_cache_root(shared=shared) == shared_cache
    else:
        assert audb.default_cache_root(shared=shared) == cache


def test_empty_shared_cache(shared_cache):
    # Handle non-existing cache folder
    # See https://github.com/audeering/audb/issues/125
    audeer.rmdir(shared_cache)
    assert not os.path.exists(shared_cache)
    df = audb.cached(shared=True)
    assert len(df) == 0
    # Handle empty shared cache folder
    # See https://github.com/audeering/audb/issues/126
    audeer.mkdir(shared_cache)
    df = audb.cached(shared=True)
    assert "name" in df.columns


def test_cached_name(cache):
    # Empty cache
    df = audb.cached()
    assert len(df) == 0
    df = audb.cached(name=DB_NAMES[0])
    assert len(df) == 0
    # Load first database
    audb.load(DB_NAMES[0], verbose=False)
    df = audb.cached()
    assert len(df) == 1
    assert set(df["name"]) == {DB_NAMES[0]}
    df = audb.cached(name=DB_NAMES[0])
    assert len(df) == 1
    assert set(df["name"]) == {DB_NAMES[0]}
    df = audb.cached(name=DB_NAMES[1])
    assert len(df) == 0
    # Load second database
    audb.load(DB_NAMES[1], verbose=False)
    df = audb.cached()
    assert len(df) == 2
    assert set(df["name"]) == set(DB_NAMES)
    df = audb.cached(name=DB_NAMES[0])
    assert len(df) == 1
    assert set(df["name"]) == {DB_NAMES[0]}
    df = audb.cached(name=DB_NAMES[1])
    assert len(df) == 1
    assert set(df["name"]) == {DB_NAMES[1]}
    df = audb.cached(name="non-existent")
    assert len(df) == 0



import os
import sys
import threading
import time

import pytest

import audbackend
import audeer
import audformat.testing

import audb


DB_NAME = "test_lock"


class SlowFileSystem(audbackend.backend.FileSystem):
    r"""Emulate a slow file system.

    Introduces a short delay when getting a file from the backend.
    This ensures that timeouts are reached in the tests.

    """

    def _get_file(self, *args):
        time.sleep(0.1)
        super()._get_file(*args)


audb.Repository.register("slow-file-system", SlowFileSystem)


class CrashFileSystem(audbackend.backend.FileSystem):
    r"""Emulate a file system that crashes.

    Raises an exception when getting a file from the backend.

    """

    def _get_file(self, *args):
        raise RuntimeError()


audb.Repository.register("crash-file-system", CrashFileSystem)


def lock_paths(cache):
    r"""Return list of lock file locations."""
    paths = []
    for version in audb.versions(DB_NAME):
        paths.append(
            audeer.path(
                cache,
                DB_NAME,
                version,
                ".lock",
            )
        )
        paths.append(
            audeer.path(
                cache,
                DB_NAME,
                version,
                audb.Flavor().short_id,
                ".lock",
            )
        )
    return paths


@pytest.fixture(
    scope="function",
    autouse=True,
)
def assert_lock_file_is_deleted():
    r"""Tests if all lock files are deleted."""
    assert not any(
        [os.path.exists(path) for path in lock_paths(audb.default_cache_root())]
    )
    yield
    assert not any(
        [os.path.exists(path) for path in lock_paths(audb.default_cache_root())]
    )


@pytest.fixture(
    scope="function",
    autouse=True,
)
def set_repositories(persistent_repository, request):
    r"""Access module wide repository with custom backends."""
    audb.config.REPOSITORIES = [
        audb.Repository(
            name=persistent_repository.name,
            host=persistent_repository.host,
            backend=request.param,
        ),
    ]


@pytest.fixture(
    scope="module",
    autouse=True,
)
def dbs(tmpdir_factory, persistent_repository):
    r"""Publish databases.

    This publishes a database with the name ``DB_NAME``
    and the versions 1.0.0 and 2.0.0
    to a module wide repository.

    """
    db_root = tmpdir_factory.mktemp("db")

    # create db

    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME
    db.schemes["scheme"] = audformat.Scheme()
    audformat.testing.add_table(
        db,
        "table",
        "filewise",
        num_files=[0, 1, 2],
    )
    db.attachments["file"] = audformat.Attachment("extra/file.txt")
    db.attachments["folder"] = audformat.Attachment("extra/folder")
    audeer.mkdir(db_root, "extra/folder/sub-folder")
    for file in [
        "extra/file.txt",
        "extra/folder/file1.txt",
        "extra/folder/file2.txt",
        "extra/folder/sub-folder/file3.txt",
    ]:
        with open(audeer.path(db_root, file), "w") as fp:
            fp.write("Some text")
    db.save(db_root)
    audformat.testing.create_audio_files(db)

    # publish 1.0.0

    audb.publish(
        db_root,
        "1.0.0",
        persistent_repository,
        verbose=False,
    )

    # publish 2.0.0

    audformat.testing.add_table(
        db,
        "empty",
        "filewise",
        num_files=0,
    )
    db.save(db_root)
    audb.publish(
        db_root,
        "2.0.0",
        persistent_repository,
        previous_version="1.0.0",
        verbose=False,
    )


def load_deps():
    return audb.dependencies(
        DB_NAME,
        version="1.0.0",
    )


@pytest.mark.parametrize(
    "set_repositories",
    ["slow-file-system"],
    indirect=True,
)
@pytest.mark.parametrize(
    "multiprocessing",
    [
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    "num_workers",
    [
        10,
    ],
)
def test_lock_dependencies(
    set_repositories,
    multiprocessing,
    num_workers,
):
    # avoid
    # AttributeError: module pytest has no attribute CACHE_ROOT
    # when multiprocessing=True on Windows and macOS
    if multiprocessing and sys.platform in ["win32", "darwin"]:
        return

    result = audeer.run_tasks(
        load_deps,
        [([], {})] * num_workers,
        num_workers=num_workers,
        multiprocessing=multiprocessing,
    )

    assert len(result) == num_workers


def load_header():
    return audb.info.header(
        DB_NAME,
        version="1.0.0",
    )


@pytest.mark.parametrize(
    "set_repositories",
    ["slow-file-system"],
    indirect=True,
)
@pytest.mark.parametrize(
    "multiprocessing",
    [
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    "num_workers",
    [
        10,
    ],
)
def test_lock_header(set_repositories, multiprocessing, num_workers):
    # avoid
    # AttributeError: module pytest has no attribute CACHE_ROOT
    # when multiprocessing=True on Windows and macOS
    if multiprocessing and sys.platform in ["win32", "darwin"]:
        return

    result = audeer.run_tasks(
        load_header,
        [([], {})] * num_workers,
        num_workers=num_workers,
        multiprocessing=multiprocessing,
    )

    assert len(result) == num_workers


def load_db(timeout):
    return audb.load(
        DB_NAME,
        version="1.0.0",
        timeout=timeout,
        verbose=False,
    )


@pytest.mark.parametrize(
    "set_repositories",
    ["slow-file-system"],
    indirect=True,
)
@pytest.mark.parametrize(
    "multiprocessing",
    [
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    "num_workers, timeout, expected",
    [
        (2, -1, 2),
        (2, 9999, 2),
        (2, 0, 1),
    ],
)
def test_lock_load(
    set_repositories,
    multiprocessing,
    num_workers,
    timeout,
    expected,
):
    # avoid
    # AttributeError: module pytest has no attribute CACHE_ROOT
    # when multiprocessing=True on Windows and macOS
    if multiprocessing and sys.platform in ["win32", "darwin"]:
        return

    warns = not multiprocessing and num_workers != expected
    params = [([timeout], {})] * num_workers
    if warns:
        with pytest.warns(
            UserWarning,
            match=audb.core.define.TIMEOUT_MSG,
        ):
            result = audeer.run_tasks(
                load_db,
                params,
                num_workers=num_workers,
                multiprocessing=multiprocessing,
            )
    else:
        result = audeer.run_tasks(
            load_db,
            params,
            num_workers=num_workers,
            multiprocessing=multiprocessing,
        )
    result = [x for x in result if x is not None]

    assert len(result) == expected


@pytest.mark.parametrize(
    "set_repositories",
    ["crash-file-system"],
    indirect=True,
)
def test_lock_load_crash(set_repositories):
    with pytest.raises(audbackend.BackendError):
        load_db(-1)


@pytest.mark.parametrize(
    "set_repositories",
    ["file-system"],
    indirect=True,
)
def test_lock_load_from_cached_versions(
    persistent_repository,
    set_repositories,
):
    # ensure immediate timeout if cache folder is locked
    cached_version_timeout = audb.core.define.CACHED_VERSIONS_TIMEOUT
    audb.core.define.CACHED_VERSIONS_TIMEOUT = 0

    # load version 1.0.0
    db_v1 = audb.load(
        DB_NAME,
        version="1.0.0",
        verbose=False,
    )

    # load new files added in version 2.0.0
    audb.load(
        DB_NAME,
        version="2.0.0",
        tables="empty",
        verbose=False,
    )

    # switch to crash backend to ensure remaining files
    # must be copied from version 1.0.0
    audb.config.REPOSITORIES = [
        audb.Repository(
            name=persistent_repository.name,
            host=persistent_repository.host,
            backend="crash-file-system",
        ),
    ]

    # lock cache folder of version 1.0.0
    def lock_v1():
        with audb.core.lock.FolderLock(db_v1.root):
            event.wait()

    event = threading.Event()
    thread = threading.Thread(target=lock_v1)
    thread.start()

    # -> loading missing table from cache fails
    with pytest.raises(audbackend.BackendError):
        audb.load(
            DB_NAME,
            version="2.0.0",
            tables="table",
            only_metadata=True,
            verbose=False,
        )

    # release cache folder of version 1.0.0
    event.set()
    thread.join()

    # -> loading missing table from cache succeeds
    audb.load(
        DB_NAME,
        version="2.0.0",
        tables="table",
        only_metadata=True,
        verbose=False,
    )

    # lock cache folder of version 1.0.0
    event.clear()
    thread = threading.Thread(target=lock_v1)
    thread.start()

    # -> loading missing media from cache fails
    with pytest.raises(audbackend.BackendError):
        audb.load(
            DB_NAME,
            version="2.0.0",
            verbose=False,
        )

    # release cache folder of version 1.0.0
    event.set()
    thread.join()

    # -> loading missing media from cache succeeds
    audb.load(
        DB_NAME,
        version="2.0.0",
        verbose=False,
    )

    # reset timeout
    audb.core.define.CACHED_VERSIONS_TIMEOUT = cached_version_timeout


def load_attachment():
    return audb.load_attachment(
        DB_NAME,
        "folder",
        version="1.0.0",
        verbose=False,
    )


@pytest.mark.parametrize(
    "set_repositories",
    ["slow-file-system"],
    indirect=True,
)
@pytest.mark.parametrize(
    "multiprocessing",
    [
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    "num_workers",
    [
        4,
    ],
)
def test_lock_load_attachment(
    set_repositories,
    multiprocessing,
    num_workers,
):
    # Avoid
    # AttributeError: module pytest has no attribute CACHE_ROOT
    # when multiprocessing=True on Windows and macOS
    if multiprocessing and sys.platform in ["win32", "darwin"]:
        return

    result = audeer.run_tasks(
        load_attachment,
        [([], {})] * num_workers,
        num_workers=num_workers,
        multiprocessing=multiprocessing,
    )

    assert len(result) == num_workers


def load_media(timeout):
    return audb.load_media(
        DB_NAME,
        "audio/001.wav",
        version="1.0.0",
        timeout=timeout,
        verbose=False,
    )


@pytest.mark.parametrize(
    "set_repositories",
    ["slow-file-system"],
    indirect=True,
)
@pytest.mark.parametrize(
    "multiprocessing",
    [
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    "num_workers, timeout, expected",
    [
        (2, -1, 2),
        (2, 9999, 2),
        (2, 0, 1),
    ],
)
def test_lock_load_media(
    set_repositories,
    multiprocessing,
    num_workers,
    timeout,
    expected,
):
    # avoid
    # AttributeError: module pytest has no attribute CACHE_ROOT
    # when multiprocessing=True on Windows and macOS
    if multiprocessing and sys.platform in ["win32", "darwin"]:
        return

    warns = not multiprocessing and num_workers != expected
    params = [([timeout], {})] * num_workers
    if warns:
        with pytest.warns(
            UserWarning,
            match=audb.core.define.TIMEOUT_MSG,
        ):
            result = audeer.run_tasks(
                load_media,
                params,
                num_workers=num_workers,
                multiprocessing=multiprocessing,
            )
    else:
        result = audeer.run_tasks(
            load_media,
            params,
            num_workers=num_workers,
            multiprocessing=multiprocessing,
        )
    result = [x for x in result if x is not None]

    assert len(result) == expected


def load_table():
    return audb.load_table(
        DB_NAME,
        "table",
        version="1.0.0",
        verbose=False,
    )


@pytest.mark.parametrize(
    "set_repositories",
    ["slow-file-system"],
    indirect=True,
)
@pytest.mark.parametrize(
    "multiprocessing",
    [
        False,
        True,
    ],
)
@pytest.mark.parametrize(
    "num_workers",
    [
        10,
    ],
)
def test_lock_load_table(set_repositories, multiprocessing, num_workers):
    # avoid
    # AttributeError: module pytest has no attribute CACHE_ROOT
    # when multiprocessing=True on Windows and macOS
    if multiprocessing and sys.platform in ["win32", "darwin"]:
        return

    result = audeer.run_tasks(
        load_table,
        [([], {})] * num_workers,
        num_workers=num_workers,
        multiprocessing=multiprocessing,
    )

    assert len(result) == num_workers



import os

import pytest

import audeer
import audformat

import audb


@pytest.fixture(scope="function", autouse=False)
def same_cache():
    r"""Set shared cache to same folder as user cache."""
    current_shared_cache = audb.config.SHARED_CACHE_ROOT
    audb.config.SHARED_CACHE_ROOT = audb.config.CACHE_ROOT

    yield

    audb.config.SHARED_CACHE_ROOT = current_shared_cache


def test_loading_attachments_from_cache(tmpdir, repository, same_cache):
    r"""Ensures loading attachments from cache works.

    As reported in https://github.com/audeering/audb/issues/314,
    loading an attachment for a database
    fails to acquire a lock,
    if the database has a previous version in the cache,
    and ``audb.config.CACHE_ROOT`` and ``audb.config.SHARED_CACHE_ROOT``
    point to the same folder.

    """
    # Create version 1.0.0 of database,
    # publish and load
    db_name = "db"
    db_version = "1.0.0"
    db_root = audeer.mkdir(audeer.path(tmpdir, db_name))
    db = audformat.Database(db_name)
    db.description = f"Version {db_version} of database."
    filename = "file.txt"
    with open(audeer.path(db_root, filename), "w") as file:
        file.write(f"{filename}\n")
    db.attachments[filename] = audformat.Attachment(filename)
    db.save(db_root)

    audb.publish(db_root, db_version, repository)

    db = audb.load(db_name, version=db_version, verbose=False)

    # Create version 2.0.0 of database,
    # and publish and load
    db = audb.load_to(db_root, db_name, version=db_version)
    db_version = "2.0.0"
    db.description = f"Version {db_version} of database."
    db.save(db_root)

    audb.publish(db_root, db_version, repository)

    db = audb.load(db_name, version=db_version, verbose=False)

    # Ensure attachment file is loaded
    assert list(db.attachments) == [filename]
    for attachment in list(db.attachments):
        for file in db.attachments[attachment].files:
            assert os.path.exists(audeer.path(db.root, file))
